import inspect
import os
import re
import pathlib
from typing import List

import pytest

from . import utils
from .conftest import FuzzyCallChecker
from .deployment_utils import UniversumRunner
from .utils import python, python_version


@pytest.fixture(name='runner_with_analyzers')
def fixture_runner_with_analyzers(docker_main: UniversumRunner):
    docker_main.environment.install_python_module("pylint")
    docker_main.environment.install_python_module("mypy")
    docker_main.environment.assert_successful_execution("apt install -y uncrustify clang-format")
    yield docker_main


class ConfigData:
    def __init__(self):
        self.text = "from universum.configuration_support import Configuration, Step\n"
        self.text += "configs = Configuration()\n"

    def add_cmd(self, name: str, cmd: str, step_config: str = '') -> 'ConfigData':
        step_config = ', ' + step_config if step_config else ''
        self.text +=\
            f"configs += Configuration([Step(name='{name}', command={cmd}{step_config})])\n"
        return self

    def add_analyzer(self, analyzer: str, arguments: List[str], step_config: str = '') -> 'ConfigData':
        name = f"Run {analyzer}"
        args = [f", '{arg}'" for arg in arguments]
        cmd = f"['{python()}', '-m', 'universum.analyzers.{analyzer}'{''.join(args)}]"
        step_config = ', ' + step_config if step_config else ''
        step_config = 'code_report=True' + step_config
        return self.add_cmd(name, cmd, step_config)

    def finalize(self) -> str:
        return inspect.cleandoc(self.text)


source_code_python = """
"Docstring."

s: str = "world"
print(f"Hello {s}.")
"""

source_code_c = """
int main() {
\treturn 0;
}
"""

json_report_minimal = """
[]
"""

json_report = """
[
    {
        "path": "my_path/my_file",
        "message": "Error!",
        "symbol": "testSymbol",
        "line": 1
    }
]
"""

sarif_report_minimal = """
{
  "version": "2.1.0",
  "runs": [
    {
      "tool": { "driver": { "name": "Dummy" } },
      "results": [ ]
    }
  ]
}
"""

sarif_report = """
{
  "$schema": "https://raw.githubusercontent.com/oasis-tcs/sarif-spec/master/Schemata/sarif-schema-2.1.0.json",
  "version": "2.1.0",
  "runs": [
    {
      "tool": {
        "driver": {
          "name": "Checkstyle",
          "semanticVersion": "8.43",
          "version": "8.43"
        }
      },
      "results": [
        {
          "level": "warning",
          "locations": [
            {
              "physicalLocation": {
                "artifactLocation": {
                  "uri": "my_path/my_file"
                },
                "region": {
                  "startColumn": 1,
                  "startLine": 1
                }
              }
            }
          ],
          "message": {
            "text": "Error!"
          },
          "ruleId": "testRule"
        }
      ]
    }
  ]
}
"""

config_uncrustify = """
code_width = 120
input_tab_size = 2
"""

config_clang_format = """
---
AllowShortFunctionsOnASingleLine: Empty
"""

log_fail = r'Found [0-9]+ issues'
log_success = r'Issues not found'


@pytest.mark.parametrize('tested_contents, expected_success', [
    [[json_report_minimal], True],
    [[json_report], False],
    [[sarif_report_minimal], True],
    [[sarif_report], False],
    [[json_report_minimal, sarif_report_minimal], True],
    [[json_report, sarif_report], False],
    [[json_report_minimal, sarif_report], False],
    [[json_report, sarif_report_minimal], False],
], ids=[
    'json_no_issues',
    'json_issues_found',
    'sarif_no_issues',
    'sarif_issues_found',
    'both_tested_no_issues',
    'both_tested_issues_in_both',
    'both_tested_issues_in_sarif',
    'both_tested_issues_in_json',
])
def test_code_report_direct_log(runner_with_analyzers: UniversumRunner, tested_contents, expected_success):
    config = ConfigData()
    step_config = "code_report=True"
    for idx, tested_content in enumerate(tested_contents):
        prelim_report = "report_file_" + str(idx)
        full_report = "${CODE_REPORT_FILE}"
        (runner_with_analyzers.local.root_directory / prelim_report).write_text(tested_content)
        config.add_cmd("Report " + str(idx), f"[\"bash\", \"-c\", \"cat './{prelim_report}' >> '{full_report}'\"]",
                       step_config)
    log = runner_with_analyzers.run(config.finalize())
    expected_log = log_success if expected_success else log_fail
    assert re.findall(expected_log, log), f"'{expected_log}' is not found in '{log}'"


@pytest.mark.parametrize('analyzers, extra_args, tested_content, expected_success', [
    [['uncrustify'], [], source_code_c, True],
    [['uncrustify'], [], source_code_c.replace('\t', ' '), False],    # by default uncrustify converts spaces to tabs
    [['clang_format'], [], source_code_c.replace('\t', '  '), True],  # by default clang-format expands tabs to 2 spaces
    [['clang_format'], [], source_code_c.replace('\t', ' '), False],
    [['clang_format', 'uncrustify'], [], source_code_c.replace('\t', ' '), False],
    [['pylint', 'mypy'], ["--python-version", python_version()], source_code_python, True],
    [['pylint'], ["--python-version", python_version()], source_code_python + '\n', False],
    [['mypy'], ["--python-version", python_version()], source_code_python.replace(': str', ': int'), False],
    [['pylint', 'mypy'], ["--python-version", python_version()], source_code_python.replace(': str', ': int') + '\n', False],
    # TODO: add test with rcfile
    # TODO: parametrize test for different versions of python
], ids=[
    'uncrustify_no_issues',
    'uncrustify_found_issues',
    'clang_format_no_issues',
    'clang_format_found_issues',
    'clang_format_and_uncrustify_found_issues',
    'pylint_and_mypy_both_no_issues',
    'pylint_found_issues',
    'mypy_found_issues',
    'pylint_and_mypy_found_issues_independently',
])
def test_code_report_log(runner_with_analyzers: UniversumRunner, analyzers, extra_args, tested_content, expected_success):
    common_args = [
        "--result-file", "${CODE_REPORT_FILE}",
        "--files", "source_file",
    ]
    (runner_with_analyzers.local.root_directory / "source_file").write_text(tested_content)
    config = ConfigData()
    for analyzer in analyzers:
        args = common_args + extra_args
        if analyzer == 'uncrustify':
            args += ["--cfg-file", "cfg"]
            (runner_with_analyzers.local.root_directory / "cfg").write_text(config_uncrustify)
        elif analyzer == 'clang_format':
            (runner_with_analyzers.local.root_directory / ".clang-format").write_text(config_clang_format)

        config.add_analyzer(analyzer, args)

    log = runner_with_analyzers.run(config.finalize())
    expected_log = log_success if expected_success else log_fail
    assert re.findall(expected_log, log), f"'{expected_log}' is not found in '{log}'"
    if not expected_success:
        for analyzer in analyzers:  # confirm that all analyzers fail independently
            assert re.findall(fr'Run {analyzer} - [^\n]*Failed', log), f"'{analyzer}' info is not found in '{log}'"


def test_without_code_report_command(runner_with_analyzers: UniversumRunner):
    log = runner_with_analyzers.run(utils.simple_test_config)
    pattern = re.compile(f"({log_fail}|{log_success})")
    assert not pattern.findall(log)


@pytest.mark.parametrize('analyzer', ['pylint', 'mypy', 'uncrustify', 'clang_format'])
@pytest.mark.parametrize('arg_set, expected_log', [
    [["--files", "source_file.py"], "error: the following arguments are required: --result-file"],
    [["--files", "source_file.py", "--result-file"], "result-file: expected one argument"],
    [["--result-file", "${CODE_REPORT_FILE}"], "error: the following arguments are required: --files"],
    [["--files", "--result-file", "${CODE_REPORT_FILE}"], "files: expected at least one argument"],
])
def test_analyzer_common_params(runner_with_analyzers: UniversumRunner, analyzer, arg_set, expected_log):
    test_analyzer_specific_params(runner_with_analyzers, analyzer, arg_set, expected_log)


@pytest.mark.parametrize('analyzer', ['pylint', 'mypy'])
@pytest.mark.parametrize('arg_set, expected_log', [
    [["--python-version", "--files", "source_file.py", "--result-file", "${CODE_REPORT_FILE}"],
     "python-version: expected one argument"],
])
def test_analyzer_python_version_params(runner_with_analyzers: UniversumRunner, analyzer, arg_set, expected_log):
    test_analyzer_specific_params(runner_with_analyzers, analyzer, arg_set, expected_log)


@pytest.mark.parametrize('analyzer, arg_set, expected_log', [
    ['pylint', ["--python-version", python_version(), "--files", "source_file",
                "--result-file", "${CODE_REPORT_FILE}", '--rcfile'],
     "rcfile: expected one argument"],
    ['uncrustify', ["--files", "source_file", "--result-file", "${CODE_REPORT_FILE}"],
     "Please specify the '--cfg-file' parameter or set 'UNCRUSTIFY_CONFIG' environment variable"],
    ['uncrustify', ["--files", "source_file", "--result-file", "${CODE_REPORT_FILE}",
                    "--cfg-file", "cfg", "--output-directory", "."],
     "Target folder must not be identical to source folder"],
    ['clang_format', ["--files", "source_file", "--result-file", "${CODE_REPORT_FILE}", "--output-directory", "."],
     "Target folder must not be identical to source folder"],
])
def test_analyzer_specific_params(runner_with_analyzers: UniversumRunner, analyzer, arg_set, expected_log):
    source_file = runner_with_analyzers.local.root_directory / "source_file"
    source_file.write_text(source_code_python)

    log = runner_with_analyzers.run(ConfigData().add_analyzer(analyzer, arg_set).finalize())
    assert re.findall(fr'Run {analyzer} - [^\n]*Failed', log), f"'{analyzer}' info is not found in '{log}'"
    assert expected_log in log, f"'{expected_log}' is not found in '{log}'"


@pytest.mark.parametrize('analyzer, extra_args, tested_content, expected_success, expected_artifact', [
    ['uncrustify', ["--report-html"], source_code_c, True, False],
    ['uncrustify', ["--report-html"], source_code_c.replace('\t', ' '), False, True],
    ['uncrustify', [], source_code_c.replace('\t', ' '), False, False],
    ['clang_format', ["--report-html"], source_code_c.replace('\t', '  '), True, False],
    ['clang_format', ["--report-html"], source_code_c, False, True],
    ['clang_format', [], source_code_c, False, False],
], ids=[
    "uncrustify_html_file_not_needed",
    "uncrustify_html_file_saved",
    "uncrustify_html_file_disabled",
    "clang_format_html_file_not_needed",
    "clang_format_html_file_saved",
    "clang_format_html_file_disabled",
])
def test_diff_html_file(runner_with_analyzers: UniversumRunner, analyzer,
                        extra_args, tested_content, expected_success, expected_artifact):

    root = runner_with_analyzers.local.root_directory
    source_file = root / "source_file"
    source_file.write_text(tested_content)
    common_args = [
        "--result-file", "${CODE_REPORT_FILE}",
        "--files", "source_file",
        "--output-directory", "diff_temp"
    ]
    if analyzer == 'uncrustify':
        (root / "cfg").write_text(config_uncrustify)
        common_args.extend(["--cfg-file", "cfg"])
    elif analyzer == 'clang_format':
        (root / ".clang-format").write_text(config_clang_format)


    args = common_args + extra_args
    extra_config = "artifacts='./diff_temp/source_file.html'"
    log = runner_with_analyzers.run(ConfigData().add_analyzer(analyzer, args, extra_config).finalize())

    expected_log = log_success if expected_success else log_fail
    assert re.findall(expected_log, log), f"'{expected_log}' is not found in '{log}'"
    expected_artifacts_state = "Success" if expected_artifact else "Failed"
    expected_log = f"Collecting artifacts for the 'Run {analyzer}' step - [^\n]*{expected_artifacts_state}"
    assert re.findall(expected_log, log), f"'{expected_log}' is not found in '{log}'"


def test_code_report_extended_arg_search(tmp_path: pathlib.Path, stdout_checker: FuzzyCallChecker):
    """
    Test if ${CODE_REPORT_FILE} is replaced not only in --result-file argument of the Step
    """
    env = utils.LocalTestEnvironment(tmp_path, "main")
    env.settings.Vcs.type = "none"
    env.settings.LocalMainVcs.source_dir = str(tmp_path)

    source_file = tmp_path / "source_file.py"
    source_file.write_text(source_code_python + '\n')

    config = f"""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Run static pylint", code_report=True, artifacts="${{CODE_REPORT_FILE}}", command=[
    'bash', '-c', 'cd "{os.getcwd()}" && {python()} -m universum.analyzers.pylint --result-file="${{CODE_REPORT_FILE}}" \
                   --python-version {python_version()} --files {str(source_file)}'])])
"""

    env.configs_file.write_text(config)

    env.run()
    stdout_checker.assert_has_calls_with_param(log_fail, is_regexp=True)
    assert os.path.exists(os.path.join(env.settings.ArtifactCollector.artifact_dir, "Run_static_pylint.json"))


def test_code_report_extended_arg_search_embedded(tmp_path: pathlib.Path, stdout_checker: FuzzyCallChecker):
    env = utils.LocalTestEnvironment(tmp_path, "main")
    env.settings.Vcs.type = "none"
    env.settings.LocalMainVcs.source_dir = str(tmp_path)

    source_file = tmp_path / "source_file.py"
    source_file.write_text(source_code_python + '\n')

    config = """
from universum.configuration_support import Configuration, Step

configs = Configuration([Step(critical=True)]) * Configuration([
    Step(name='This is step', command=["ls"]),
    Step(name='This is step to unfold', code_report=True, report_artifacts='${CODE_REPORT_FILE}',
         command=['bash', '-c', 'echo ${CODE_REPORT_FILE}']),
])
"""

    env.configs_file.write_text(config)

    env.run()
    stdout_checker.assert_absent_calls_with_param("${CODE_REPORT_FILE}")



import os
import signal
import subprocess
import time
import shutil
import pathlib
from typing import Any

import pytest

from .deployment_utils import UniversumRunner, LocalSources
from .utils import python, simple_test_config


def get_line_with_text(text: str, log: str) -> str:
    for line in log.splitlines():
        if text in line:
            return line
    return ""


def test_minimal_execution(docker_main_and_nonci: UniversumRunner):
    log = docker_main_and_nonci.run(simple_test_config)
    assert docker_main_and_nonci.local.repo_file.name in log


def test_artifacts(docker_main: UniversumRunner):
    config = """
from universum.configuration_support import Configuration

mkdir = Configuration([dict(name="Create directory", command=["mkdir", "-p"])])
mkfile = Configuration([dict(name="Create file", command=["touch"])])
dirs1 = Configuration([dict(name=" one/two{}/three".format(str(x)),
                         command=["one/two{}/three".format(str(x))]) for x in range(0, 6)])
files1 = Configuration([dict(name=" one/two{}/three/file{}.txt".format(str(x), str(x)),
                          command=["one/two{}/three/file{}.txt".format(str(x), str(x))])
                     for x in range(0, 6)])

dirs2 = Configuration([dict(name=" one/three", command=["one/three"])])
files2 = Configuration([dict(name=" one/three/file.sh", command=["one/three/file.sh"])])

artifacts = Configuration([dict(name="Existing artifacts", artifacts="one/**/file*", report_artifacts="one/*"),
                           dict(name="Missing report artifacts", report_artifacts="non_existing_file"),
                           dict(name="Missing all artifacts", artifacts="something", report_artifacts="something_else")])

configs = mkdir * dirs1 + mkdir * dirs2 + mkfile * files1 + mkfile * files2 + artifacts
    """
    log = docker_main.run(config)
    assert 'Failed' in get_line_with_text("Collecting artifacts for the 'Missing all artifacts' step - ", log)
    assert 'Success' in get_line_with_text("Collecting artifacts for the 'Missing report artifacts' step - ", log)

    assert os.path.exists(os.path.join(docker_main.artifact_dir, "three.zip"))
    assert os.path.exists(os.path.join(docker_main.artifact_dir, "two2.zip"))
    assert os.path.exists(os.path.join(docker_main.artifact_dir, "file1.txt"))
    assert os.path.exists(os.path.join(docker_main.artifact_dir, "file.sh"))


def test_background_steps(docker_main_and_nonci: UniversumRunner):
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

background = Configuration([dict(name="Background", background=True)])
sleep = Configuration([dict(name=' long step', command=["sleep", "1"])])
multiply = Configuration([dict(name="_1"), dict(name="_2"), dict(name="_3")])
wait = Configuration([dict(name='Step requiring background results',
                      command=["echo", "test passed"], finish_background=True)])

script = Configuration([dict(name=" unsuccessful step", command=["ls", "non-existent-file"])])

configs = background * (script + sleep * multiply) + wait + background * (sleep + script)
""")
    assert "All ongoing background steps should be finished before next step execution" in log
    assert 'Failed' in get_line_with_text("Background unsuccessful step - ", log)

    # Test background after failed foreground (regression)
    docker_main_and_nonci.clean_artifacts()
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Bad step", command=["ls", "not_a_file"]),
                         dict(name="Good bg step", command=["touch", "file"],
                         background=True, artifacts="file")])
""")
    assert "All ongoing background steps completed" in log
    assert os.path.exists(os.path.join(docker_main_and_nonci.artifact_dir, "file"))

    # Test TC step failing
    docker_main_and_nonci.clean_artifacts()
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Bad bg step", command=["ls", "not_a_file"], background=True)])
""", additional_parameters=" -ot tc")
    assert "##teamcity[buildProblem description" in log

    # Test multiple failing background steps
    docker_main_and_nonci.clean_artifacts()
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Bad step 1", command=["ls", "not_a_file"], background=True),
                         dict(name="Bad step 2", command=["ls", "not_a_file"], background=True)])
""")
    assert 'Failed' in get_line_with_text("Bad step 2 - ", log)


def test_critical_steps(docker_main_and_nonci: UniversumRunner):
    # Test linear
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Good step", command=["echo", "step succeeded"]),
                         dict(name="Bad step", command=["ls", "not_a_file"], critical=True),
                         dict(name="Extra step", command=["echo", "This shouldn't be in log."])])
""")
    assert "'Extra step' skipped because of critical step failure" in log
    assert "This shouldn't be in log." not in log

    # Test embedded: critical step, critical substep
    docker_main_and_nonci.clean_artifacts()
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

upper = Configuration([dict(name="Group 1"),
                       dict(name="Group 2", critical=True),
                       dict(name="Group 3")])

lower = Configuration([dict(name=", step 1", command=["echo", "step succeeded"]),
                       dict(name=", step 2", command=["ls", "not_a_file"], critical=True),
                       dict(name=", step 3", command=["echo", "This shouldn't be in log."])])

configs = upper * lower
""")
    assert "'Group 3, step 1' skipped because of critical step failure" in log
    assert "'Group 2, step 1' skipped because of critical step failure" not in log

    # Test embedded: critical step, non-critical substep
    docker_main_and_nonci.clean_artifacts()
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

upper = Configuration([dict(name="Group 1", critical=True),
                       dict(name="Group 2")])

lower = Configuration([dict(name=", step 1", command=["echo", "step succeeded"]),
                       dict(name=", step 2", command=["ls", "not_a_file"]),
                       dict(name=", step 3", command=["echo", "This should be in log."])])

configs = upper * lower
""")
    assert "'Group 2, step 1' skipped because of critical step failure" in log
    assert "This should be in log." in log

    # Test critical non-commands
    docker_main_and_nonci.clean_artifacts()
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Group 1")])

configs *= Configuration([dict(name=", step 1", command=["echo", "step succeeded"]),
                          dict(name=", step 2", command=["this-is-not-a-command"], critical=True),
                          dict(name=", step 3", command=["echo", "This shouldn't be in log."])])

configs += Configuration([dict(name="Linear non-command", command=["this-is-not-a-command"], critical=True),
                          dict(name="Extra step", command=["echo", "This shouldn't be in log."])])
""")
    assert "This shouldn't be in log." not in log

    # Test successful critical step after failing non-critical step
    docker_main_and_nonci.clean_artifacts()
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Group 1")])

configs *= Configuration([dict(name=", step 1", command=["echo", "step succeeded"]),
                          dict(name=", step 2", command=["this-is-not-a-command"]),
                          dict(name=", step 3", command=["echo", "This should be in log 1."], critical=True),
                          dict(name=", step 4", command=["echo", "This should be in log 2."])])
    """)
    assert "This should be in log 1." in log
    assert "This should be in log 2." in log

    # Test background
    docker_main_and_nonci.clean_artifacts()
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

group = Configuration([dict(name="Group")])

subgroup1 = Configuration([dict(name=" 1, step 1", command=["ls", "not_a_file"], background=True),
                           dict(name=" 1, step 2", command=["echo", "This should be in log - 1"],
                             finish_background=True)])

subgroup2 = Configuration([dict(name=" 2, step 1", command=["ls", "not_a_file"], critical=True,
                                background=True),
                           dict(name=" 2, step 2", command=["echo", "This should be in log - 2"])])

subgroup3 = Configuration([dict(name=" 3, step 1", command=["echo", "This shouldn't be in log."],
                                finish_background=True),
                           dict(name=" 3, step 2", command=["echo", "This shouldn't be in log."])])

configs = group * subgroup1 + group * subgroup2 + group * subgroup3
configs += Configuration([dict(name="Additional step", command=["echo", "This should be in log - 3"])])
""")
    assert "This shouldn't be in log." not in log
    assert "This should be in log - 1" in log
    assert "This should be in log - 2" in log
    assert "This should be in log - 3" in log


def test_empty_steps(docker_main_and_nonci: UniversumRunner):
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration, Step

configs = Configuration([Step(name="Step one"),
                         Step(name="Step two", critical=True),
                         Step(name="Step three", background=True)])
""")
    assert "'RunningStep' object has no attribute 'process'" not in log
    assert "Nothing was executed: this background step had no command" in log


def test_minimal_git(docker_main_with_vcs: UniversumRunner):
    log = docker_main_with_vcs.run(simple_test_config, vcs_type="git")
    assert docker_main_with_vcs.git.repo_file.name in log


def test_minimal_p4(docker_main_with_vcs: UniversumRunner):
    log = docker_main_with_vcs.run(simple_test_config, vcs_type="p4")
    assert docker_main_with_vcs.perforce.repo_file.name in log


def test_p4_params(docker_main_with_vcs: UniversumRunner):
    p4 = docker_main_with_vcs.perforce.p4
    p4_file = docker_main_with_vcs.perforce.repo_file
    config = f"""
from universum.configuration_support import Configuration, Step

configs = Configuration([Step(name="Test step", command=["cat", "{p4_file.name}"])])
"""

    # Prepare SYNC_CHANGELIST
    sync_cl = p4.run_changes("-s", "submitted", "-m1", docker_main_with_vcs.perforce.depot)[0]["change"]
    p4.run_edit(docker_main_with_vcs.perforce.depot)
    p4_file.write_text("This line shouldn't be in file.\n")
    change = p4.fetch_change()
    change["Description"] = "Rename basic config"
    p4.run_submit(change)

    # Prepare SHELVE_CHANGELIST
    p4.run_edit(docker_main_with_vcs.perforce.depot)
    p4_file.write_text("This line should be in file.\n")
    change = p4.fetch_change()
    change["Description"] = "CL for shelving"
    shelve_cl = p4.save_change(change)[0].split()[1]
    p4.run_shelve("-fc", shelve_cl)

    # Do not pass params
    log = docker_main_with_vcs.run(config, vcs_type="p4")
    assert "This line shouldn't be in file." in log
    assert "This line should be in file." not in log

    # Pass params via command line
    docker_main_with_vcs.clean_artifacts()
    log = docker_main_with_vcs.run(config, vcs_type="p4",
                                   additional_parameters=" -p4h=" + sync_cl + " -p4s=" + shelve_cl)
    assert "This line shouldn't be in file." not in log
    assert "This line should be in file." in log

    # Pass params via environment variables
    docker_main_with_vcs.clean_artifacts()
    log = docker_main_with_vcs.run(config, vcs_type="p4", environment=["SYNC_CHANGELIST=" + sync_cl,
                                                                       "SHELVE_CHANGELIST_1=" + shelve_cl])
    assert "This line shouldn't be in file." not in log
    assert "This line should be in file." in log


def empty_required_params_ids(param: Any) -> str:
    if isinstance(param, bool):  # url_error_expected
        return 'negative' if param else 'positive'
    return str(param)


@pytest.mark.parametrize('url_error_expected, parameters, env', [
    [True, "", []],
    [True, " -ssu=''", []],
    [True, "", ["SWARM_SERVER="]],
    [True, " --build-only-latest -ssu=''", []],

    # negative Test cases
    [False, " -ssu=http://swarm", []],
    [False, "", ["SWARM_SERVER=http://swarm"]],
    [False, " --build-only-latest -ssu=http://swarm", []]
], ids=empty_required_params_ids)
def test_empty_required_params(docker_main_with_vcs: UniversumRunner, url_error_expected, parameters, env):
    url_error = "URL of the Swarm server is not specified"

    log = docker_main_with_vcs.run(simple_test_config, vcs_type="p4", expected_to_fail=True,
                                   additional_parameters=" --report-to-review" + parameters, environment=env)
    if url_error_expected:
        assert url_error in log
    else:
        assert url_error not in log


def test_environment(docker_main_and_nonci: UniversumRunner):
    script = docker_main_and_nonci.local.root_directory / "script.sh"
    script.write_text("""#!/bin/bash
echo ${SPECIAL_TESTING_VARIABLE}
""")
    script.chmod(0o777)

    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Test configuration", command=["script.sh"],
                              environment={"SPECIAL_TESTING_VARIABLE": "This string should be in log"})])
""")
    assert "This string should be in log" in log

    docker_main_and_nonci.clean_artifacts()
    log = docker_main_and_nonci.run("""
from universum.configuration_support import Configuration

upper = Configuration([dict(name="Test configuration",
                            environment={"SPECIAL_TESTING_VARIABLE": "This string should be in log"})])
lower = Configuration([dict(name=" 1", command=["script.sh"], environment={"OTHER_VARIABLE": "Something"}),
                       dict(name=" 2", command=["ls", "-la"])])
configs = upper * lower
""")
    assert "This string should be in log" in log


@pytest.mark.parametrize("terminate_type", [signal.SIGINT, signal.SIGTERM], ids=["interrupt", "terminate"])
def test_abort(local_sources: LocalSources, tmp_path: pathlib.Path, terminate_type):
    config = """
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Long step", command=["sleep", "10"])]) * 5
"""
    config_file = tmp_path / "configs.py"
    config_file.write_text(config)

    with subprocess.Popen([python(), "-m", "universum",
                           "-o", "console", "-st", "local", "-vt", "none",
                           "-pr", str(tmp_path / "project_root"),
                           "-ad", str(tmp_path / "artifacts"),
                           "-fsd", str(local_sources.root_directory),
                           "-cfg", str(config_file)]) as process:
        time.sleep(5)
        process.send_signal(terminate_type)
        assert process.wait(5) == 3


def test_exit_code(local_sources: LocalSources, tmp_path: pathlib.Path):
    config = """
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Unsuccessful step", command=["exit", "1"])])
"""
    config_file = tmp_path / "configs.py"
    config_file.write_text(config)

    with subprocess.Popen([python(), "-m", "universum",
                           "-o", "console", "-st", "local", "-vt", "none",
                           "-pr", str(tmp_path / "project_root"),
                           "-ad", str(tmp_path / "artifacts"),
                           "-fsd", str(local_sources.root_directory),
                           "-cfg", str(config_file)]) as process:

        assert process.wait() == 0

    artifacts_dir = tmp_path / "artifacts"
    shutil.rmtree(str(artifacts_dir))
    with subprocess.Popen([python(), "-m", "universum", "--fail-unsuccessful",
                           "-o", "console", "-st", "local", "-vt", "none",
                           "-pr", str(tmp_path / "project_root"),
                           "-ad", str(tmp_path / "artifacts"),
                           "-fsd", str(local_sources.root_directory),
                           "-cfg", str(config_file)]) as process:
        assert process.wait() == 1



# pylint: disable = redefined-outer-name

from universum.lib.gravity import construct_component
from universum.lib.module_arguments import ModuleArgumentParser
from universum.modules.error_state import HasErrorState

text = "This is text\nwith some line breaks\n"
error_message = "This is some missing argument error message"

script = f"""
from universum.lib.module_arguments import ModuleArgumentParser
from universum.lib.gravity import construct_component
from universum.modules.error_state import HasErrorState

parser = ModuleArgumentParser()
parser.add_argument('--argument', '-a', dest='HasErrorState.argument', metavar='ARGUMENT')

namespace = parser.parse_args()
module = construct_component(HasErrorState, namespace)
print(module.read_and_check_multiline_option('argument', "{error_message}"))
if module.is_in_error_state():
    print(module.global_error_state.get_errors()[0])
"""


def construct_test_module(args):
    argument_parser = ModuleArgumentParser()
    argument_parser.add_argument('--argument', '-a', dest='HasErrorState.argument', metavar='ARGUMENT')
    settings = argument_parser.parse_args(args)
    return construct_component(HasErrorState, settings)


def test_success_multiline_variable():
    module = construct_test_module(['-a', text])
    assert module.read_and_check_multiline_option('argument', error_message) == text


def test_multiline_variable_files(tmp_path):
    var_path = tmp_path / "variable.txt"
    var_path.write_text(text)
    module = construct_test_module(['-a', f"@{str(var_path)}"])
    assert module.read_and_check_multiline_option('argument', error_message) == text

    module = construct_test_module(['-a', "@this-is-not-a-file"])
    module.read_and_check_multiline_option('argument', error_message)
    assert module.is_in_error_state()
    error = module.global_error_state.get_errors()[0]
    assert "argument" in str(error)
    assert "this-is-not-a-file" in str(error)



# pylint: disable = redefined-outer-name

import pathlib
import pytest

from . import utils
from .perforce_utils import P4TestEnvironment, PerforceWorkspace


@pytest.fixture()
def p4_submit_environment(perforce_workspace: PerforceWorkspace, tmp_path: pathlib.Path):
    yield P4TestEnvironment(perforce_workspace, tmp_path, test_type="submit")


def test_fail_changing_non_checked_out_file(p4_submit_environment: P4TestEnvironment):
    target_file = p4_submit_environment.vcs_client.nonwritable_file
    text = utils.randomize_name("This is change ")
    with pytest.raises(IOError) as excinfo:
        with open(str(target_file), "w", encoding="utf-8") as tmpfile:
            tmpfile.write(text + "\n")

    assert "Permission denied" in str(excinfo.value)


def test_success_changing_checked_out_file(p4_submit_environment: P4TestEnvironment):
    target_file = p4_submit_environment.vcs_client.nonwritable_file

    p4_submit_environment.vcs_client.p4.run("edit", str(target_file))

    text = utils.randomize_name("This is change ")
    target_file.write_text(text + "\n")

    change = p4_submit_environment.vcs_client.p4.run_change("-o")[0]
    change["Description"] = "Test submit"
    p4_submit_environment.vcs_client.p4.run_submit(change)

    assert p4_submit_environment.vcs_client.file_present(str(target_file))



import os
import re
from datetime import datetime
from re import Match
from typing import Optional, List

from ansi2html import Ansi2HTMLConverter

from .base_output import BaseOutput

__all__ = [
    "HtmlOutput"
]


class HtmlOutput(BaseOutput):
    default_name = "universum_log.html"

    def __init__(self, *args, log_name: str = default_name, **kwargs):
        super().__init__(*args, **kwargs)
        self._log_name: str = log_name
        self._log_path: Optional[str] = None
        self.artifact_dir_ready: bool = False
        self._log_buffer: List[str] = []
        self._block_level: int = 0
        self.module_dir: str = os.path.dirname(os.path.realpath(__file__))
        self.ansi_converter: Ansi2HTMLConverter = Ansi2HTMLConverter(inline=True, escaped=False)

    def log_execution_start(self, title: str, version: str) -> None:
        head_content: str = self._build_html_head()
        html_header: str = f"<!DOCTYPE html><html><head>{head_content}</head><body>"
        html_header += '<input type="checkbox" id="dark-checkbox"><label for="dark-checkbox"></label>'
        html_header += '<input type="checkbox" id="time-checkbox"><label for="time-checkbox"></label>'
        html_header += '<pre>'

        self._log_buffered(html_header)
        self.log(self._build_execution_start_msg(title, version))

    def log_execution_finish(self, title: str, version: str) -> None:
        self.log(self._build_execution_finish_msg(title, version))
        html_footer: str = "</pre>"
        with open(os.path.join(self.module_dir, "html_output.js"), encoding="utf-8") as js_file:
            html_footer += f"<script>{js_file.read()}</script>"
        html_footer += "</body></html>"
        self._log_line(html_footer)

    def log(self, line: str) -> None:
        self._log_line(f"==> {line}")

    def log_error(self, description: str) -> None:
        self._log_line(f'<span class="exceptionTag">Error:</span> {description}')

    def log_external_command(self, command: str) -> None:
        self._log_line(f"$ {command}")

    def log_stdout(self, line: str) -> None:
        self._log_line(line)

    def log_stderr(self, line: str) -> None:
        self._log_line(f'<span class="stderrTag">stderr:</span> {line}')

    def open_block(self, num_str: str, name: str) -> None:
        opening_html: str = f'<input type="checkbox" id="{num_str}" class="hide"/>' + \
                            f'<label for="{num_str}"><span class="sectionLbl">'
        closing_html: str = "</span></label><div>"
        name_html: str = f'<span class="sectionTitle">{name}</span>'
        self._log_line(f"{opening_html}{num_str} {name_html}{closing_html}", with_line_separator=False)
        self._block_level += 1

    def close_block(self, num_str: str, name: str, status: str) -> None:
        self._block_level -= 1
        indent: str = "  " * self._block_level
        closing_html: str = '</div><span class="nl"></span>'
        status_html: str = f'<span class="{status.lower()}Status">[{status}]</span>'
        self._log_line(f"{indent} \u2514 {status_html}{closing_html}", with_line_separator=False)
        self._log_line("")

    def log_skipped(self, message: str) -> None:
        self._log_line(f'<span class="skipped">{message}</span>')

    def log_summary_step(self, step_title: str, has_children: bool, status: str) -> None:
        if not has_children:
            step_title += f' - <span class="{status.lower()}Status">{status}</span>'
        self.log(step_title)

    def report_build_problem(self, description: str) -> None:
        raise RuntimeError("Html output doesn't support reporting build problem.")

    def set_build_status(self, status: str) -> None:
        raise RuntimeError("Html output doesn't support setting build title.")

    def set_artifact_dir(self, artifact_dir: str) -> None:
        self._log_path = os.path.join(artifact_dir, self._log_name)

    def _log_line(self, line: str, with_line_separator: bool = True) -> None:
        if with_line_separator and not line.endswith(os.linesep):
            line += os.linesep
        self._log_buffered(self._build_time_stamp() + self._build_indent() + line)

    def _log_buffered(self, line: str) -> None:
        line = self._wrap_links(line)
        line = self._ansi_codes_to_html(line)
        if not self.artifact_dir_ready:
            self._log_buffer.append(line)
            return
        if self._log_buffer:
            self._log_and_clear_buffer()
        self._write_to_file(line)

    def _log_and_clear_buffer(self) -> None:
        for buffered_line in self._log_buffer:
            self._write_to_file(buffered_line)
        self._log_buffer = []

    def _write_to_file(self, line: str) -> None:
        if not self._log_path:
            raise RuntimeError("Artifact directory was not set")

        with open(self._log_path, "a", encoding="utf-8") as file:
            file.write(line)

    def _build_indent(self) -> str:
        indent_str: List[str] = []
        for x in range(0, self._block_level):
            indent_str.append("  " * x)
            indent_str.append(" |   ")
        return "".join(indent_str)

    def _build_html_head(self) -> str:
        head: List[str] = ['<meta content="text/html;charset=utf-8" http-equiv="Content-Type">',
                           '<meta content="utf-8" http-equiv="encoding">']
        with open(os.path.join(self.module_dir, "html_output.css"), encoding="utf-8") as css_file:
            head.append(f"<style>{css_file.read()}</style>")
        return "".join(head)

    def _ansi_codes_to_html(self, line: str) -> str:
        return self.ansi_converter.convert(line, full=False)

    @staticmethod
    def _wrap_links(line: str) -> str:
        position_shift: int = 0
        pattern = r"(?:http|https|ftp|file|mailto):(?:\\ |\S)+"
        for match in re.finditer(pattern, line):
            link: str = match.group()
            wrapped_link: str = f'<a href="{link}">{link}</a>'
            link_start_pos: int = match.start() + position_shift
            link_end_pos: int = match.end() + position_shift
            line = line[:link_start_pos] + wrapped_link + line[link_end_pos:]
            position_shift += len(wrapped_link) - len(link)
        return line

    @staticmethod
    def _build_time_stamp() -> str:
        now = datetime.now()
        return now.astimezone().strftime('<span class="time" title="%Z (UTC%z)">%Y-%m-%d %H:%M:%S</span> ')



# pylint: disable = redefined-outer-name

import os
import shutil
import pathlib
import pytest

from universum import __main__
from .conftest import FuzzyCallChecker
from .perforce_utils import P4TestEnvironment, PerforceWorkspace


@pytest.fixture()
def perforce_environment(perforce_workspace: PerforceWorkspace, tmp_path: pathlib.Path):
    yield P4TestEnvironment(perforce_workspace, tmp_path, test_type="main")


def test_p4_forbidden_local_revert(perforce_environment: P4TestEnvironment, stdout_checker: FuzzyCallChecker):
    p4 = perforce_environment.vcs_client.p4

    config = """
from universum.configuration_support import Configuration

configs = Configuration([dict(name="Restrict changes", command=["chmod", "-R", "555", "."]),
                         dict(name="Check", command=["ls", "-la"])])
"""

    perforce_environment.shelve_config(config)
    result = __main__.run(perforce_environment.settings)
    # Clean up the directory at once to make sure it doesn't remain non-writable even if some assert fails
    os.system(f"chmod -R 777 {perforce_environment.temp_dir}")
    shutil.rmtree(str(perforce_environment.temp_dir))

    assert result == 0

    stdout_checker.assert_has_calls_with_param("[Errno 13] Permission denied")

    # make sure there are no pending CLs in the workspace
    assert not p4.run_changes("-c", perforce_environment.client_name, "-s", "pending")
    # make sure there are no pending changes in default CL
    assert not p4.run_opened("-C", perforce_environment.client_name)


def test_p4_print_exception_before_run(perforce_environment: P4TestEnvironment, stdout_checker: FuzzyCallChecker):
    p4 = perforce_environment.vcs_client.p4
    client = p4.fetch_client(perforce_environment.client_name)
    client["Options"] = "noallwrite noclobber nocompress locked nomodtime normdir"
    p4.save_client(client)

    settings = perforce_environment.settings
    result = __main__.run(settings)

    # Update client at once to make sure it doesn't remain locked even if some assert fails
    client = p4.fetch_client(perforce_environment.client_name)
    client["Options"] = "noallwrite noclobber nocompress unlocked nomodtime normdir"
    p4.save_client(client)

    assert result != 0
    stdout_checker.assert_has_calls_with_param(
        f"Errors during command execution( \"p4 client -d {perforce_environment.client_name}\" )")


def test_p4_print_exception_in_finalize(perforce_environment: P4TestEnvironment, stdout_checker: FuzzyCallChecker):
    p4 = perforce_environment.vcs_client.p4
    client = p4.fetch_client(perforce_environment.client_name)
    client["Options"] = "noallwrite noclobber nocompress locked nomodtime normdir"
    p4.save_client(client)

    settings = perforce_environment.settings
    settings.Main.finalize_only = True
    result = __main__.run(settings)

    # Update client at once to make sure it doesn't remain locked even if some assert fails
    client = p4.fetch_client(perforce_environment.client_name)
    client["Options"] = "noallwrite noclobber nocompress unlocked nomodtime normdir"
    p4.save_client(client)

    assert result == 0
    stdout_checker.assert_has_calls_with_param(
        f"Errors during command execution( \"p4 client -d {perforce_environment.client_name}\" )")
    stdout_checker.assert_has_calls_with_param("[Errno 2] No such file or directory")


@pytest.mark.parametrize('cl_list', [["132,456"], ["@123,@456"], ["//depot/...@,//depot2/...@"],
                                     ["//depot/...,//depot2/..."], ["132", "456"], ["@123", "4@456"],
                                     ["//depot/...@", "//depot2/...@"], ["//depot/...", "//depot2/..."]])
def test_p4_print_exception_in_sync(perforce_environment: P4TestEnvironment, stdout_checker: FuzzyCallChecker, cl_list):
    perforce_environment.settings.PerforceMainVcs.sync_cls = cl_list
    perforce_environment.run(expect_failure=True)
    text = f"Something went wrong when processing sync CL parameter ('{str(cl_list)}')"
    stdout_checker.assert_has_calls_with_param(text)


def test_p4_print_exception_wrong_shelve(perforce_environment: P4TestEnvironment, stdout_checker: FuzzyCallChecker):
    cl = perforce_environment.vcs_client.make_a_change()
    perforce_environment.settings.PerforceMainVcs.shelve_cls = [cl]

    # This is not the 'already committed' case of Swarm review, so it actually should fail
    perforce_environment.run(expect_failure=True)
    stdout_checker.assert_has_calls_with_param(
        f"Errors during command execution( \"p4 unshelve -s {cl} -f\" )")
    stdout_checker.assert_has_calls_with_param(f"[Error]: 'Change {cl} is already committed.'")



from typing import cast, Any, Callable, ClassVar, Dict, Generic, List, Optional, Type, TypeVar, Union

from typing_extensions import Protocol

__all__ = [
    "Module",
    "Dependency",
    "construct_component",
    "define_arguments_recursive"
]


class ModuleSettings:
    def __init__(self, cls, main_settings) -> None:
        object.__setattr__(self, "cls", cls)
        object.__setattr__(self, "main_settings", main_settings)

    def __getattribute__(self, item):
        cls = object.__getattribute__(self, "cls")
        main_settings = object.__getattribute__(self, "main_settings")
        for entry in cls.__mro__:
            try:
                settings = getattr(main_settings, entry.__name__)
                return getattr(settings, item)
            except AttributeError:
                continue

        raise AttributeError("'" + cls.__name__ + "' object has no setting '" + item + "'")

    def __setattr__(self, key, value):
        cls = object.__getattribute__(self, "cls")
        main_settings = object.__getattribute__(self, "main_settings")
        for entry in cls.__mro__:
            try:
                settings = getattr(main_settings, entry.__name__)
                getattr(settings, key)
                setattr(settings, key, value)
                return
            except AttributeError:
                continue

        raise AttributeError("'" + cls.__name__ + "' object has no setting '" + key + "'")


class Settings:
    def __init__(self, cls: Type['Module']) -> None:
        self.cls: Type['Module'] = cls

    def __get__(self, obj: 'Module', objtype: Any = None) -> 'ModuleSettings':
        return ModuleSettings(self.cls, obj.main_settings)

    def __set__(self, obj: 'Module', settings: 'Settings'):
        setattr(obj.main_settings, self.cls.__name__, settings)


class HasModulesMapping(Protocol):
    active_modules: Dict[Type['Module'], 'Module']


class Module:
    settings: ClassVar['Settings']
    main_settings: 'HasModulesMapping'

    def __init__(self, *args, **kwargs):  # workaround for type checking
        pass

    def __new__(cls: Type['Module'], main_settings: 'HasModulesMapping', *args, **kwargs) -> 'Module':
        instance: 'Module' = super(Module, cls).__new__(cls)
        instance.main_settings = main_settings
        return instance


ComponentTypeT = TypeVar('ComponentTypeT', bound=Module)


def construct_component(cls: Type[ComponentTypeT], main_settings: 'HasModulesMapping',
                        *args, **kwargs) -> ComponentTypeT:

    if not getattr(main_settings, "active_modules", None):
        main_settings.active_modules = {}

    if cls not in main_settings.active_modules:
        cls.settings = Settings(cls)

        # The module instance is constructed manually, because we need to set
        # settings attribute before constructor is called
        instance: 'Module' = cls.__new__(cls, main_settings=main_settings)

        # https://github.com/python/mypy/blob/master/mypy/checkmember.py#180
        # Accessing __init__ in statically typed code would compromise
        # type safety unless used via super().
        # noinspection PyArgumentList
        # pylint: disable = unnecessary-dunder-call
        instance.__init__(*args, **kwargs)  # type: ignore
        main_settings.active_modules[cls] = instance
    return cast(ComponentTypeT, main_settings.active_modules[cls])


DependencyTypeT = TypeVar('DependencyTypeT', bound=Module)


class Dependency(Generic[DependencyTypeT]):
    def __init__(self, cls: Type[DependencyTypeT]) -> None:
        self.cls = cls

    # The source and the target of the dependency are different modules,
    # so we need to use different type variables to annotate them.
    # instance parameter of the __get__ method is source module.
    def __get__(self, instance: Module, owner: Any) -> Callable[..., DependencyTypeT]:
        def constructor_function(*args, **kwargs) -> DependencyTypeT:
            return construct_component(self.cls, instance.main_settings, *args, **kwargs)
        return constructor_function


def get_dependencies(cls: Type[Module], result: Optional[List[Type[Module]]] = None) -> List[Type[Module]]:
    if result is None:
        result = []

    result.append(cls)

    # parent classes
    all_dependencies: List[Type[Module]]
    all_dependencies = [x for x in cls.__mro__ if issubclass(x, Module) and x != Module and x != cls]
    # dependencies
    all_dependencies.extend([x.cls for x in cls.__dict__.values() if isinstance(x, Dependency)])

    for dependency in all_dependencies:
        if dependency not in result:
            get_dependencies(dependency, result)

    return result


def define_arguments_recursive(cls, argument_parser):  # TODO: return here after ModuleArgumentParser is annotated
    modules = get_dependencies(cls)

    for current_module in modules:
        if "define_arguments" in current_module.__dict__:
            argument_parser.dest_prefix = current_module.__name__ + '.'
            # noinspection PyUnresolvedReferences
            current_module.define_arguments(argument_parser)

    argument_parser.dest_prefix = ''



import argparse
import difflib
import pathlib
import shutil
from typing import Callable, List, Optional, Tuple

from . import utils


class HtmlDiffFileWriter:

    def __init__(self, target_folder: pathlib.Path, wrapcolumn: int, tabsize: int) -> None:
        self.target_folder = target_folder
        self.differ = difflib.HtmlDiff(tabsize=tabsize, wrapcolumn=wrapcolumn)

    def __call__(self, file: pathlib.Path, src: List[str], target: List[str]) -> None:
        file_relative = file.relative_to(pathlib.Path.cwd())
        out_file_name: str = str(file_relative).replace('/', '_') + '.html'
        with open(self.target_folder.joinpath(out_file_name), 'w', encoding="utf-8") as out_file:
            out_file.write(self.differ.make_file(src, target, context=False))


DiffWriter = Callable[[pathlib.Path, List[str], List[str]], None]


def diff_analyzer_output_parser(files: List[Tuple[pathlib.Path, pathlib.Path]],
                                write_diff_file: Optional[DiffWriter]
                                ) -> List[utils.ReportData]:
    result: List[utils.ReportData] = []
    for src_file, dst_file in files:
        with open(src_file, encoding="utf-8") as src:
            src_lines = src.readlines()
        with open(dst_file, encoding="utf-8") as fixed:
            fixed_lines = fixed.readlines()

        issues = _get_issues_from_diff(src_file, src_lines, fixed_lines)
        if issues and write_diff_file:
            write_diff_file(src_file, src_lines, fixed_lines)
        result.extend(issues)
    return result


def _get_issues_from_diff(src_file: pathlib.Path, src: List[str], target: List[str]) -> List[utils.ReportData]:
    result = []
    matching_blocks: List[difflib.Match] = \
        difflib.SequenceMatcher(a=src, b=target).get_matching_blocks()
    previous_match = matching_blocks[0]
    for match in matching_blocks[1:]:
        block = _get_mismatching_block(previous_match, match, src, target)
        previous_match = match
        if not block:
            continue
        line, before, after = block
        path: pathlib.Path = src_file.relative_to(pathlib.Path.cwd())
        message = _get_issue_message(before, after)
        result.append(utils.ReportData(
            symbol="Code Style issue",
            message=message,
            path=str(path),
            line=line
        ))

    return result


def _get_issue_message(before: str, after: str) -> str:
    # The maximum number of lines to write separate comments for
    # If exceeded, summarized comment will be provided instead
    max_lines = 11
    diff_size = len(before.splitlines())
    if diff_size > max_lines:
        message = f"\nLarge block of code ({diff_size} lines) has issues\n" + \
                  f"Non-compliant code blocks exceeding {max_lines} lines are not reported\n"
    else:
        # Message with before&after
        message = f"\nOriginal code:\n```diff\n{before}```\n" + \
                  f"Fixed code:\n```diff\n{after}```\n"
    return message


def _get_mismatching_block(previous_match: difflib.Match,  # src[a:a+size] = target[b:b+size]
                           match: difflib.Match,
                           src: List[str], target: List[str]
                           ) -> Optional[Tuple[int, str, str]]:
    previous_match_end_in_src = previous_match.a + previous_match.size
    previous_match_end_in_target = previous_match.b + previous_match.size
    match_start_in_src = match.a
    match_start_in_target = match.b
    if previous_match_end_in_src == match_start_in_src:
        return None
    line = match_start_in_src
    before = _get_text_for_block(previous_match_end_in_src - 1, match_start_in_src, src)
    after = _get_text_for_block(previous_match_end_in_target - 1, match_start_in_target, target)
    return line, before, after


def _get_text_for_block(start: int, end: int, lines: List[str]) -> str:
    return _replace_whitespace_characters(''.join(lines[start: end]))


_whitespace_character_mapping = {
    " ": "\u00b7",
    "\t": "\u2192\u2192\u2192\u2192",
    "\n": "\u2193\u000a"
}.items()


def _replace_whitespace_characters(line: str) -> str:
    for old_str, new_str in _whitespace_character_mapping:
        line = line.replace(old_str, new_str)
    return line


def diff_analyzer_argument_parser(description: str, module_path: str, output_directory: str) -> argparse.ArgumentParser:
    parser = utils.create_parser(description, module_path)
    parser.add_argument("--output-directory", "-od", dest="output_directory", default=output_directory,
                        help=f"Directory to store fixed files and HTML files with diff; the default "
                             f"value is '{output_directory}'. Has to be distinct from source directory")
    parser.add_argument("--report-html", dest="write_html", action="store_true", default=False,
                        help="(optional) Set to generate html reports for each modified file")
    return parser


def diff_analyzer_common_main(settings: argparse.Namespace) -> None:
    settings.target_folder = utils.normalize_path(settings.output_directory)
    if settings.target_folder.exists() and settings.target_folder.samefile(pathlib.Path.cwd()):
        raise EnvironmentError("Target folder must not be identical to source folder")

    settings.target_folder.mkdir(parents=True, exist_ok=True)

    if not shutil.which(settings.executable):
        raise EnvironmentError(f"{settings.name} executable '{settings.executable}' is not found. "
                               f"Please install {settings.name} or fix the executable name.")



# pylint: disable=W0223
import glob
import importlib
import os
from typing import Any

from .base_vcs import BaseVcs, BaseDownloadVcs, BasePollVcs, BaseSubmitVcs
from ..error_state import HasErrorState
from ..output import HasOutput
from ..structure_handler import HasStructure
from ...lib import utils
from ...lib.ci_exception import CriticalCiException
from ...lib.utils import make_block, convert_to_str

__all__ = [
    "GitMainVcs",
    "GitSubmitVcs",
    "GitPollVcs"
]

git: Any = None


def catch_git_exception(ignore_if=None):
    return utils.catch_exception("GitCommandError", ignore_if)


class GitVcs(BaseVcs, HasOutput, HasStructure, HasErrorState):
    """
    This class contains CI functions for interaction with Git
    """

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Git", "Git repository settings")

        parser.add_argument("--git-repo", "-gr", dest="repo", metavar="GIT_REPO",
                            help="See your project home page for exact repository identifier, passed to 'git clone'. "
                                 "If using SSH, '--git-repo' format is 'ssh://user@server:port/detailed/path'")
        parser.add_argument("--git-refspec", "-grs", dest="refspec", metavar="GIT_REFSPEC",
                            help="Any additional refspec to be fetched")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        global git
        try:
            git = importlib.import_module("git")
            remote = importlib.import_module("git.remote")
        except ImportError as e:
            text = "Error: using VCS type 'git' requires official Git CLI and Python package 'gitpython' " \
                   "to be installed to the system. Please refer to `Prerequisites` chapter of project " \
                   "documentation for detailed instructions"
            raise ImportError(text) from e

        self.check_required_option("repo", """
            The git repo is not specified.

            The repo defines the location of project source codes. Please specify the git
            repo by using '--git-repo' ('-gr') command line parameter or by setting GIT_REPO
            environment variable.
            """)

        class Progress(remote.RemoteProgress):
            def __init__(self, out, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.out = out

            def line_dropped(self, line):
                self.out.log(line)

        self.repo = None
        self.logger = Progress(self.out)

        if getattr(self.settings, "refspec", None):
            if self.settings.refspec.startswith("origin/"):
                self.refspec = self.settings.refspec[7:]
            else:
                self.refspec = self.settings.refspec
        else:
            self.refspec = None

    @make_block("Cloning repository")
    @catch_git_exception()
    def clone_and_fetch(self, history_depth=None):
        self.out.log("Cloning '" + self.settings.repo + "'...")
        destination_directory = convert_to_str(self.settings.project_root)
        self._clone(history_depth, destination_directory, self.settings.repo)
        self.sources_need_cleaning = True
        self.append_repo_status("Git repo: " + self.settings.repo + "\n\n")

        self.out.log("Please note that default remote name is 'origin'")
        if self.settings.refspec:
            self.repo.remotes.origin.fetch(refspec=self.settings.refspec, progress=self.logger)
            self.append_repo_status("Fetched refspec: " + self.settings.refspec + "\n")

    @catch_git_exception()
    def _clone(self, history_depth, destination_directory, clone_url):
        if history_depth:
            self.repo = git.Repo.clone_from(clone_url, destination_directory, depth=history_depth,
                                            no_single_branch=True, progress=self.logger)
        else:
            self.repo = git.Repo.clone_from(clone_url, destination_directory, progress=self.logger)


class GitMainVcs(GitVcs, BaseDownloadVcs):
    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Git")

        parser.add_argument("--git-checkout-id", "-gco", dest="checkout_id", metavar="GIT_CHECKOUT_ID",
                            help="A commit ID to checkout. "
                                 "Could be exact commit hash, or branch name, or tag, etc.")

        parser.add_argument("--git-cherry-pick-id", "-gcp", action="append", nargs='+',
                            metavar="GIT_CHERRYPICK_ID", dest="cherrypick_id",
                            help="List of commit IDs to be cherry-picked, separated by comma. "
                                 "'--git-cherry-pick-id' can be added to the command line several times")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.checkout_id = None

    @make_block("Checking out")
    @catch_git_exception()
    def check_out(self):
        if self.settings.checkout_id:
            self.checkout_id = self.settings.checkout_id
        elif self.settings.refspec:
            self.checkout_id = "FETCH_HEAD"
        else:
            self.checkout_id = "HEAD"
        self.out.log("Checking out '" + self.checkout_id + "'...")
        self.repo.git.checkout(self.checkout_id)
        self.append_repo_status("Checked out: " + self.checkout_id + "\n")

    @make_block("Cherry-picking")
    @catch_git_exception()
    def cherry_pick(self):
        cherrypick_id_list = sorted(list(set(utils.unify_argument_list(self.settings.cherrypick_id))))
        self.append_repo_status("Cherry-picked commits:")
        for commit in cherrypick_id_list:
            self.out.log("Cherry-picking '" + commit + "'...")
            self.repo.git.cherry_pick(commit, "--no-commit")
            self.append_repo_status(" " + commit)
        self.append_repo_status("\n")

    def _diff_against_reference_commit(self, commit_id):
        """Details. Depending on a 'git' version 'rename file' operation generates
        different output. It could be a sequence of 'add' and 'delete' or single
        'rename' operation.
        """

        status_mapping = {
            'A': "add",
            'C': "copy",
            'D': "delete",
            'M': "modify",
            'R': "rename",
            'T': "change file type",
            'U': "unmerged",
            'X': "unknown"
        }
        result = []
        for line in self.repo.git.diff(commit_id, name_status=True).splitlines():
            try:
                diff_record = line.split()
                try:
                    # Some status letters are followed by score, e.g. 'R86'
                    status = status_mapping[diff_record[0][0]]
                except KeyError:
                    status = diff_record[0]

                result.append({"action": status,
                               "repo_path": diff_record[1],
                               "local_path": utils.parse_path(diff_record[-1], self.settings.project_root)})
            except IndexError:
                self.out.log_error(line)

        return result

    def calculate_file_diff(self):
        return self._diff_against_reference_commit(self.checkout_id)

    @catch_git_exception()
    def prepare_repository(self):
        self.clone_and_fetch()
        self.check_out()
        if self.settings.cherrypick_id:
            self.cherry_pick()

    def copy_cl_files_and_revert(self):
        raise RuntimeError("Git doesn't support calculating diff for code report steps.")


class GitSubmitVcs(GitVcs, BaseSubmitVcs, HasErrorState):
    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Git")

        parser.add_argument("--git-user", "-gu", dest="user", metavar="GITUSER",
                            help="Git user name for submitting")
        parser.add_argument("--git-email", "-ge", dest="email", metavar="GITEMAIL",
                            help="Git user email for submitting")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.check_required_option("user", """
            The git user name is not specified.

            Submitting changes to repository requires setting user name and email. Please
            specify the user name by using '--git-user' ('-gu') command line parameter or by
            setting GITUSER environment variable
            """)

        self.check_required_option("email", """
            The git user email is not specified.

            Submitting changes to repository requires setting user name and email. Please
            specify the user email by using '--git-email' ('-ge') command line parameter or
            by setting GITEMAIL environment variable
            """)

    def get_list_of_modified(self, file_list):
        """
        Output of 'git status --porcelain' for most cases looks as following:

             M path/file.name
            ?? path/newly/created.file
             D path/deleted.file
            R  old/path/file -> new/path/file

        And for '--edit-only' submit option we should filter the 'M' records
        :param file_list: full list of vcs and directories to be reconciled
        :return: list of corresponding modified vcs
        """
        result = []
        all_changes = self.repo.git.status(porcelain=True).splitlines()

        modified_files = set()
        for file_record in all_changes:
            record_parameters = file_record.split(" ")
            if record_parameters[-2] == "M":
                full_path = utils.parse_path(record_parameters[-1], self.settings.project_root)
                modified_files.add(full_path)

        for file_path in file_list:
            all_matches = glob.glob(file_path)
            relative_path = os.path.relpath(file_path, self.settings.project_root)
            if not all_matches:
                self.out.log(f"Skipping '{relative_path}'...")
                continue

            for matching_path in all_matches:
                relative_path = os.path.relpath(matching_path, self.settings.project_root)
                if os.path.isdir(matching_path):
                    files_in_dir = [os.path.relpath(item, self.settings.project_root)
                                    for item in modified_files if item.startswith(file_path)]
                    if not files_in_dir:
                        self.out.log(f"Skipping '{relative_path}'...")
                    result.extend(files_in_dir)
                else:
                    if matching_path in modified_files:
                        result.append(relative_path)
                    else:
                        self.out.log(f"Skipping '{relative_path}'...")
        return result

    def git_commit_locally(self, description, file_list, edit_only=False):
        try:
            self.repo = git.Repo(convert_to_str(self.settings.project_root))
        except git.exc.NoSuchPathError as e:
            raise CriticalCiException("No such directory as '" + self.settings.project_root + "'") from e
        except git.exc.InvalidGitRepositoryError as e:
            raise CriticalCiException("'" + self.settings.project_root + "' does not contain a Git repository") from e

        with self.repo.config_writer() as configurator:
            configurator.set_value("user", "name", self.settings.user)
            configurator.set_value("user", "email", self.settings.email)

        file_list = [utils.parse_path(item, self.settings.project_root) for item in file_list]
        relative_path_list = [os.path.relpath(item, self.settings.project_root) for item in file_list]

        if edit_only:
            self.repo.git.add(self.get_list_of_modified(file_list))
        else:
            self.repo.git.add(relative_path_list, all=True)

        repo_status = self.repo.git.status()
        nothing_committed = ("nothing added to commit" in repo_status or
                             "no changes added to commit" in repo_status or
                             "nothing to commit" in repo_status)
        if nothing_committed:
            return 0

        self.out.log(self.repo.git.commit(m=description))
        commit_id = str(self.repo.head.commit)
        self.out.log("Full commit ID is " + commit_id)
        return commit_id

    def submit_new_change(self, description, file_list, review=False, edit_only=False):
        change = self.git_commit_locally(description, file_list, edit_only=edit_only)
        if change == 0:
            return 0

        if review:
            raise CriticalCiException("'review' commits to non-gerrit Git are not supported at the moment. "
                                      "Specify temp branch for deletable commit manually if needed")

        self.repo.remotes.origin.push(progress=self.logger)
        return change


class GitPollVcs(GitVcs, BasePollVcs):
    def get_changes(self, changes_reference=None, max_number='1'):
        self.clone_and_fetch()
        if not changes_reference:
            changes_reference = {}
        result = {}

        branch_name = self.refspec
        result[branch_name] = []

        last_change = self.repo.git.log("origin/" + branch_name, pretty="oneline", max_count=1).split(" ")[0]
        reference_change = changes_reference.get(branch_name, last_change)

        # Ranges like "commit^.." do not work for single-commit branches, so reference change is processed manually
        result[branch_name].append(reference_change)
        submitted_changes = self.repo.git.log("--first-parent", "origin/" + branch_name, reference_change + "..",
                                              pretty="oneline", max_count=max_number).splitlines()

        submitted_changes.reverse()
        for change in submitted_changes:
            result[branch_name].append(change.split(" ")[0])

        return result



# pylint: disable = redefined-outer-name

from typing import Union
import pathlib
import pytest

from universum import __main__
from .conftest import FuzzyCallChecker
from .git_utils import GitServer, GitClient, GitTestEnvironment
from .perforce_utils import PerforceWorkspace, P4TestEnvironment
from .utils import LocalTestEnvironment


def test_poll_local_vcs(tmp_path: pathlib.Path):
    env = LocalTestEnvironment(tmp_path, "poll")
    env.run()


def test_p4_success_command_line_no_changes(stdout_checker: FuzzyCallChecker,
                                            perforce_workspace: PerforceWorkspace,
                                            tmp_path: pathlib.Path):
    db_file = tmp_path / "p4poll.json"
    result = __main__.main(["poll", "-ot", "term",
                            "-vt", "p4",
                            "-f", str(db_file),
                            "-p4p", perforce_workspace.server.port,
                            "-p4u", perforce_workspace.server.user,
                            "-p4P", perforce_workspace.server.password,
                            "-p4d", perforce_workspace.depot,
                            "-jtu", "https://localhost/?%s"])
    assert result == 0
    stdout_checker.assert_has_calls_with_param("==> No changes detected")


def test_git_success_command_line_no_changes(stdout_checker: FuzzyCallChecker,
                                             git_server: GitServer,
                                             tmp_path: pathlib.Path):
    db_file = tmp_path / "gitpoll.json"
    result = __main__.main(["poll", "-ot", "term",
                            "-vt", "git",
                            "-f", str(db_file),
                            "-gr", git_server.url,
                            "-grs", git_server.target_branch,
                            "-jtu", "https://localhost/?%s"])
    assert result == 0
    stdout_checker.assert_has_calls_with_param("==> No changes detected")


def test_p4_error_command_line_wrong_port(stdout_checker: FuzzyCallChecker,
                                          perforce_workspace: PerforceWorkspace,
                                          tmp_path: pathlib.Path):
    db_file = tmp_path / "p4poll.json"
    result = __main__.main(["poll", "-ot", "term",
                            "-vt", "p4",
                            "-f", str(db_file),
                            "-p4p", "127.0.0.1:1024",
                            "-p4u", perforce_workspace.server.user,
                            "-p4P", perforce_workspace.server.password,
                            "-p4d", perforce_workspace.depot,
                            "-jtu", "https://localhost/?%s"])
    assert result != 0
    stdout_checker.assert_has_calls_with_param("TCP connect to 127.0.0.1:1024 failed.")


def test_git_error_command_line_wrong_port(stdout_checker: FuzzyCallChecker,
                                           git_server: GitServer,
                                           tmp_path: pathlib.Path):
    db_file = tmp_path / "gitpoll.json"
    result = __main__.main(["poll", "-ot", "term",
                            "-vt", "git",
                            "-f", str(db_file),
                            "-gr", "file:///non-existing-directory",
                            "-grs", git_server.target_branch,
                            "-jtu", "https://localhost/?%s"])
    assert result != 0
    stdout_checker.assert_has_calls_with_param("Cmd('git') failed due to: exit code(128)")


@pytest.fixture(params=["git", "p4"])
def poll_environment(request, perforce_workspace: PerforceWorkspace, git_client: GitClient, tmp_path: pathlib.Path):
    if request.param == "git":
        yield GitTestEnvironment(git_client, tmp_path, test_type="poll")
    else:
        yield P4TestEnvironment(perforce_workspace, tmp_path, test_type="poll")


def test_error_one_change(stdout_checker: FuzzyCallChecker, log_exception_checker: FuzzyCallChecker,
                          poll_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    # initialize working directory with initial data
    poll_environment.run_with_http_server()

    # make change in workspace
    change = poll_environment.vcs_client.make_a_change()

    # run poll again and fail triggering url because there is no server
    poll_environment.run(expect_failure=True)

    stdout_checker.assert_has_calls_with_param("==> Detected commit " + change)

    # there is no listening server
    log_exception_checker.assert_has_calls_with_param("[Errno 111] Connection refused")


def test_success_one_change(stdout_checker: FuzzyCallChecker, poll_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    # initialize working directory with initial data
    poll_environment.run_with_http_server()

    # make change in workspace
    change = poll_environment.vcs_client.make_a_change()

    collected_http = poll_environment.run_with_http_server()
    collected_http.assert_request_was_made({"cl": [change]})
    stdout_checker.assert_has_calls_with_param("==> Detected commit " + change)


def test_success_two_changes(stdout_checker: FuzzyCallChecker, poll_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    # initialize working directory with initial data
    poll_environment.run_with_http_server()

    # make changes in workspace
    change1 = poll_environment.vcs_client.make_a_change()
    change2 = poll_environment.vcs_client.make_a_change()

    # run poll again and trigger the url twice
    collected_http = poll_environment.run_with_http_server()

    stdout_checker.assert_has_calls_with_param("==> Detected commit " + change1)
    stdout_checker.assert_has_calls_with_param("==> Detected commit " + change2)

    collected_http.assert_request_was_made({"cl": [change1]})
    collected_http.assert_request_was_made({"cl": [change2]})


def test_changes_several_times(stdout_checker: FuzzyCallChecker, poll_environment: Union[GitTestEnvironment, P4TestEnvironment]):
    # initialize working directory with initial data
    poll_environment.run_with_http_server()

    # make changes in workspace
    change1 = poll_environment.vcs_client.make_a_change()
    change2 = poll_environment.vcs_client.make_a_change()

    # run poll and trigger the urls
    collected_http = poll_environment.run_with_http_server()

    stdout_checker.assert_has_calls_with_param("==> Detected commit " + change1)
    stdout_checker.assert_has_calls_with_param("==> Detected commit " + change2)

    collected_http.assert_request_was_made({"cl": [change1]})
    collected_http.assert_request_was_made({"cl": [change2]})

    # make more changes in workspace
    stdout_checker.reset()
    change3 = poll_environment.vcs_client.make_a_change()
    change4 = poll_environment.vcs_client.make_a_change()

    # run poll and trigger urls for the new changes only
    collected_http = poll_environment.run_with_http_server()

    stdout_checker.assert_has_calls_with_param("==> Detected commit " + change3)
    stdout_checker.assert_has_calls_with_param("==> Detected commit " + change4)
    stdout_checker.assert_absent_calls_with_param("==> Detected commit " + change1)
    stdout_checker.assert_absent_calls_with_param("==> Detected commit " + change2)

    collected_http.assert_request_was_made({"cl": [change3]})
    collected_http.assert_request_was_made({"cl": [change4]})
    collected_http.assert_request_was_not_made({"cl": [change1]})
    collected_http.assert_request_was_not_made({"cl": [change2]})



# -*- coding: utf-8 -*-

# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

import os
import sys
sys.path.append(os.path.abspath("./_ext/collapsible"))


# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.coverage',
              'sphinx.ext.doctest',
              'sphinx.ext.napoleon',
              'sphinx.ext.todo',
              'sphinx.ext.viewcode',
              'sphinxarg.ext',
              'collapsible']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
source_suffix = ['.rst']

# The master toctree document.
master_doc = 'index'

# General information about the project.
author = u'Ivan Keliukh'
copyright = u'Samsung Electronics Co., Ltd.'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    'logo_only': True,
    'navigation_depth': 2
}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/logo-with-text.svg'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = '_static/favicon.png'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_scaled_image_link = False


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'ci_doc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, document, class [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'ci.tex', u'CI Documentation',
     u'Ivan Keliukh', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'ci', u'CI Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'ci', u'CI Documentation',
     author, 'ci', 'Continuous Integration system.',
     'Miscellaneous'),
]

sys.path.append(os.path.dirname(__file__)+'/..')

import universum

project = universum.__title__
# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = universum.__version__
# The full version, including alpha/beta/rc tags.
release = universum.__version__

autodoc_member_order = 'bysource'
add_module_names = False


# Add replacement for Python version
# Solution from https://stackoverflow.com/a/56328457
def ultimateReplace(app, docname, source):
    result = source[0]
    for key in app.config.ultimate_replacements:
        result = result.replace(key, app.config.ultimate_replacements[key])
    source[0] = result


ultimate_replacements = {
    "{pip}": f"pip{sys.version_info.major}.{sys.version_info.minor}",
    "{python}": f"python{sys.version_info.major}.{sys.version_info.minor}"
}


def setup(app):
    app.add_config_value('ultimate_replacements', {}, True)
    app.connect('source-read', ultimateReplace)
    app.add_css_file('theme_overrides.css')



from .base_output import BaseOutput

__all__ = [
    "TeamcityOutput"
]


def escape(message: str) -> str:
    return message.replace('\r', '').replace('|', '||').replace('\'', '|\'').replace('[', '|[').replace(']', '|]')


class TeamcityOutput(BaseOutput):
    def log(self, line: str):
        print("==>", line)

    def log_error(self, description: str) -> None:
        lines = description.split("\n")
        for single_line in lines:
            print(f"##teamcity[message text='{escape(single_line)}' status='ERROR']")

    def log_external_command(self, command: str) -> None:
        print("$", command)

    def log_stdout(self, line: str) -> None:
        print(line)

    def log_stderr(self, line: str) -> None:
        lines = line.split("\n")
        for single_line in lines:
            print(f"##teamcity[message text='{escape(single_line)}' status='WARNING']")

    def open_block(self, num_str: str, name: str) -> None:
        print(f"##teamcity[blockOpened name='{num_str} {escape(name)}']")

    def close_block(self, num_str: str, name: str, status: str) -> None:
        print(f"##teamcity[blockClosed name='{num_str} {escape(name)}']")

    def log_skipped(self, message: str) -> None:
        lines = message.split("\n")
        for single_line in lines:
            print(f"##teamcity[message text='{escape(single_line)}' status='WARNING']")

    def log_summary_step(self, step_title: str, has_children: bool, status: str) -> None:
        if has_children:
            self.log(step_title)
        else:
            self.log(f"{step_title} - {status}")

    def report_build_problem(self, description: str) -> None:
        print(f"##teamcity[buildProblem description='<{escape(description)}>']")

    def set_build_status(self, status: str) -> None:
        print(f"##teamcity[buildStatus text='{escape(status)}']")



import argparse
import os
import pathlib
import re
from typing import Callable, List, Optional, Tuple

from . import utils, diff_utils


def uncrustify_argument_parser() -> argparse.ArgumentParser:
    parser = diff_utils.diff_analyzer_argument_parser("Uncrustify analyzer", __file__, "uncrustify")
    parser.add_argument("--cfg-file", "-cf", dest="cfg_file",
                        help="Name of the configuration file of Uncrustify; "
                             "can also be set via 'UNCRUSTIFY_CONFIG' env. variable")
    return parser


@utils.sys_exit
@utils.analyzer(uncrustify_argument_parser())
def main(settings: argparse.Namespace) -> List[utils.ReportData]:
    settings.name = "uncrustify"
    settings.executable = "uncrustify"
    diff_utils.diff_analyzer_common_main(settings)

    if not settings.cfg_file and 'UNCRUSTIFY_CONFIG' not in os.environ:
        raise EnvironmentError("Please specify the '--cfg-file' parameter "
                               "or set 'UNCRUSTIFY_CONFIG' environment variable")

    html_diff_file_writer: Optional[Callable[[pathlib.Path, List[str], List[str]], None]] = None
    if settings.write_html:
        wrapcolumn, tabsize = _get_wrapcolumn_tabsize(settings.cfg_file)
        html_diff_file_writer = diff_utils.HtmlDiffFileWriter(settings.target_folder, wrapcolumn, tabsize)

    cmd = ["uncrustify", "-q", "-c", settings.cfg_file, "--prefix", settings.output_directory]
    files: List[Tuple[pathlib.Path, pathlib.Path]] = []
    for src_file_absolute, target_file_absolute, src_file_relative in utils.get_files_with_absolute_paths(settings):
        files.append((src_file_absolute, target_file_absolute))
        cmd.append(src_file_relative)

    utils.run_for_output(cmd)
    return diff_utils.diff_analyzer_output_parser(files, html_diff_file_writer)


def _get_wrapcolumn_tabsize(cfg_file: str) -> Tuple[int, int]:
    wrapcolumn = 120
    tabsize = 4
    with open(cfg_file, encoding="utf-8") as config:
        for line in config.readlines():
            match = re.match(r"^\s*([A-Za-z_]+)\s*[,\=]?\s*([0-9]+)\s*$", line)
            if not match:
                continue
            groups = match.groups()
            if groups[0] == "code_width":
                wrapcolumn = int(groups[1])
            if groups[0] == "input_tab_size":
                tabsize = int(groups[1])
    return wrapcolumn, tabsize


if __name__ == "__main__":
    main()



import pathlib
import sys

from .modules.output.output import MinimalOut
from .lib.gravity import Module
__all__ = ["ConfigCreator"]

PYTHON_VERSION = f"python{sys.version_info.major}.{sys.version_info.minor}"


class ConfigCreator(Module):
    description: str = "Create a dummy project"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.out = MinimalOut()

    def execute(self) -> None:
        config_name = ".universum.py"
        self.out.log(f"Creating an example configuration file '{config_name}'")

        config = pathlib.Path(config_name)
        config.write_text(f"""#!/usr/bin/env {PYTHON_VERSION}

from universum.configuration_support import Configuration, Step

configs = Configuration([Step(name='Show directory contents', command=['ls', '-la']),
                         Step(name='Print a line', command=['bash', '-c', 'echo Hello world'])])

if __name__ == '__main__':
    print(configs.dump())
""", encoding="utf-8")
        self.out.log(f"To run with Universum, execute the following command:\n$ {PYTHON_VERSION} -m universum run")

    def finalize(self) -> None:
        pass



# pylint: disable = redefined-outer-name

import pathlib
import git
import pytest

from . import utils
from .git_utils import GitClient, GitTestEnvironment
from .perforce_utils import PerforceWorkspace, P4TestEnvironment


@pytest.fixture
def unicode_dir(tmp_path: pathlib.Path):
    unicode_dir_path = tmp_path / "Юніко́д з пробелами"
    unicode_dir_path.mkdir()
    yield unicode_dir_path


@pytest.mark.parametrize("vcs", ["git", "p4"])
@pytest.mark.parametrize("test_type", ["main", "poll", "submit"])
def test_unicode(vcs, test_type, perforce_workspace: PerforceWorkspace, git_client: GitClient, unicode_dir: pathlib.Path):
    env: utils.BaseTestEnvironment
    if vcs == "git":
        # change git client root dir to unicode path
        work_dir = unicode_dir / "client"
        work_dir.mkdir()
        git_client.repo = git.Repo.clone_from(git_client.server.url, work_dir)
        git_client.root_directory = work_dir

        env = GitTestEnvironment(git_client, unicode_dir, test_type=test_type)
    elif vcs == "p4":
        # change workspace root dir to unicode path
        root = unicode_dir / "workspace"
        root.mkdir()
        client = perforce_workspace.p4.fetch_client(perforce_workspace.client_name)
        client["Root"] = str(root)
        perforce_workspace.root_directory = root
        perforce_workspace.p4.save_client(client)
        perforce_workspace.p4.run_sync("-f", "//depot/...")

        env = P4TestEnvironment(perforce_workspace, unicode_dir, test_type=test_type)
    else:
        assert False, "Unsupported vcs type"

    if test_type == "submit":
        file_name = utils.randomize_name("new_file") + ".txt"
        temp_file = env.vcs_client.root_directory / file_name
        temp_file.write_text("This is a new file" + "\n")
        env.settings.Submit.reconcile_list = str(temp_file)

    env.run()


def test_unicode_main_local_vcs(unicode_dir: pathlib.Path):
    work_dir = unicode_dir / "local_sources"
    work_dir.mkdir()
    (work_dir / "source_file").write_text("Source file contents")

    env = utils.LocalTestEnvironment(unicode_dir, "main")
    env.settings.Vcs.type = "none"
    env.settings.LocalMainVcs.source_dir = str(work_dir)

    env.run()



import codecs
import os
import shutil
import zipfile
from typing import List, Optional, Dict, Union, TypedDict

import glob2

from ..configuration_support import Configuration, Step
from ..lib.ci_exception import CriticalCiException, CiException
from ..lib.gravity import Dependency
from ..lib.utils import make_block
from ..lib import utils
from .automation_server import AutomationServerForHostingBuild
from .output import HasOutput
from .project_directory import ProjectDirectory
from .reporter import Reporter
from .structure_handler import HasStructure
from .output.html_output import HtmlOutput


__all__ = [
    "ArtifactCollector"
]


def make_big_archive(target, source):
    save_cwd = os.getcwd()
    if source is not None:
        target = os.path.abspath(target)
        os.chdir(source)
    base_dir = os.curdir

    try:
        filename = target + ".zip"
        archive_dir = os.path.dirname(target)

        if archive_dir and not os.path.exists(archive_dir):
            os.makedirs(archive_dir)

        with zipfile.ZipFile(filename, "w", compression=zipfile.ZIP_DEFLATED,
                             allowZip64=True) as zf:
            path = os.path.normpath(base_dir)
            zf.write(path, path)
            for dirpath, dirnames, filenames in os.walk(base_dir):
                for name in sorted(dirnames):
                    path = os.path.normpath(os.path.join(dirpath, name))
                    zf.write(path, path)
                for name in filenames:
                    path = os.path.normpath(os.path.join(dirpath, name))
                    if os.path.isfile(path):
                        zf.write(path, path)
    finally:
        if source is not None:
            os.chdir(save_cwd)

    return filename


class ArtifactInfo(TypedDict):
    path: str
    clean: bool


class ArtifactCollector(ProjectDirectory, HasOutput, HasStructure):
    reporter_factory = Dependency(Reporter)
    automation_server_factory = Dependency(AutomationServerForHostingBuild)
    html_output_factory = Dependency(HtmlOutput)

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Artifact collection",
                                                     "Parameters of archiving and collecting of build artifacts")

        parser.add_argument("--artifact-dir", "-ad", dest="artifact_dir", metavar="ARTIFACT_DIR",
                            help="Directory to collect artifacts to. Default is 'artifacts'")

        parser.add_argument("--no-archive", action="store_true", dest="no_archive",
                            help="By default all directories noted as artifacts are copied as .zip archives. "
                                 "This option turn archiving off to copy bare directories to artifact directory")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reporter = self.reporter_factory()
        self.automation_server = self.automation_server_factory()

        # Needed because of wildcards
        self.collected_report_artifacts = set()

        self.file_list = set()
        self.artifact_dir = self.settings.artifact_dir

        if not self.artifact_dir:
            self.artifact_dir = os.path.join(os.getcwd(), "artifacts")
        if not os.path.exists(self.artifact_dir):
            os.makedirs(self.artifact_dir)

        self.html_output = self.html_output_factory()
        self.html_output.set_artifact_dir(self.artifact_dir)
        self.html_output.artifact_dir_ready = False

    def make_file_name(self, name):
        return utils.calculate_file_absolute_path(self.artifact_dir, name)

    # TODO: using codecs is legacy from Python2; this function needs to be refactored
    def create_text_file(self, name):
        try:
            file_name = self.make_file_name(name)
            if file_name not in self.file_list:
                if os.path.exists(file_name):
                    text = "File '" + os.path.basename(file_name) + "' already exists in artifact directory." + \
                           "\nPossible reason of this error: previous build artifacts are not cleaned"
                    raise CriticalCiException(text)

            self.file_list.add(file_name)
            file_path = self.automation_server.artifact_path(self.artifact_dir, os.path.basename(file_name))
            self.out.log("Adding file " + file_path + " to artifacts...")
            return codecs.open(file_name, "a", encoding="utf-8")          # pylint: disable = consider-using-with

        except IOError as e:
            raise CiException("The following error occurred while working with file: " + str(e)) from e

    def preprocess_artifact_list(self, artifact_list, ignore_already_existing=False):
        """
        Check artifacts for existence; remove if required; raise exception otherwise; sort and remove duplicates
        :param artifact_list: list of dictionaries with keys 'path' (for path) and 'clean' to clean it before build
        :param ignore_already_existing: will not check existence of artifacts when set to 'True'
        :return: sorted list of checked paths (including duplicates and wildcards)
        """
        for item in artifact_list:
            # Check existence in place: wildcards applied
            matches = glob2.glob(item["path"])
            if matches:
                if item["clean"]:
                    for matching_path in matches:
                        try:
                            os.remove(matching_path)  # TODO: use shutil by default
                        except OSError as e:
                            if "Is a directory" not in e.strerror:
                                raise
                            shutil.rmtree(matching_path)
                        self.out.log(f"Cleaned up '{matching_path}'")
                elif not ignore_already_existing:
                    text = "Build artifacts, such as"
                    for matching_path in matches:
                        text += f"\n * '{os.path.basename(matching_path)}'"
                    text += f"\nalready exist in '{os.path.dirname(item['path'])}' directory."
                    text += "\nPossible reason of this error: previous build results in working directory"
                    raise CriticalCiException(text)

            # Check existence in 'artifacts' directory: wildcards NOT applied
            artifact_file = os.path.join(self.artifact_dir, os.path.basename(item["path"]))
            self._check_artifact_absent(artifact_file)

            if not self.settings.no_archive:
                artifact_zip_archive = os.path.join(artifact_file + ".zip")
                self._check_artifact_absent(artifact_zip_archive)

    @make_block("Preprocessing artifact lists")
    def set_and_clean_artifacts(self, project_configs: Configuration, ignore_existing_artifacts: bool = False) -> None:
        self.html_output.artifact_dir_ready = True
        artifact_list: List[ArtifactInfo] = []
        for configuration in project_configs.all():
            if configuration.artifacts:
                self.append_config_artifact_if_present(artifact_list, configuration)
            if configuration.report_artifacts:
                self.append_config_artifact_if_present(artifact_list, configuration, is_report_artifact=True)
            if configuration.is_conditional:
                artifact_list.extend(self.get_conditional_step_branches_artifacts(configuration))

        if artifact_list:
            name = "Setting and preprocessing artifacts according to configs"
            with self.structure.block(block_name=name, pass_errors=True):
                self.preprocess_artifact_list(artifact_list, ignore_existing_artifacts)

    def get_conditional_step_branches_artifacts(self, conditional_step: Step) -> List[ArtifactInfo]:
        steps_to_process: List[Step] = []
        if conditional_step.if_succeeded:
            steps_to_process.extend(list(conditional_step.if_succeeded.all()))
        if conditional_step.if_failed:
            steps_to_process.extend(list(conditional_step.if_failed.all()))

        artifacts: List[ArtifactInfo] = []
        for step in steps_to_process:
            self.append_config_artifact_if_present(artifacts, step)
            self.append_config_artifact_if_present(artifacts, step, is_report_artifact=True)

        defined_artifacts: List[ArtifactInfo] = [artifact for artifact in artifacts if artifact]
        return defined_artifacts

    def append_config_artifact_if_present(self,
                                          artifacts: List[ArtifactInfo],
                                          step: Step,
                                          is_report_artifact: bool = False) -> None:
        artifact: Optional[ArtifactInfo] = self.get_config_artifact(step, is_report_artifact)
        if artifact:
            artifacts.append(artifact)

    def get_config_artifact(self, step: Step, is_report_artifact: bool = False) -> Optional[ArtifactInfo]:
        artifact: str = step.report_artifacts if is_report_artifact else step.artifacts
        if not artifact:
            return None
        path: str = utils.parse_path(artifact, self.settings.project_root)
        return dict(path=path, clean=step.artifact_prebuild_clean)

    def move_artifact(self, path, is_report=False):
        self.out.log("Processing '" + path + "'")
        matches = glob2.glob(path)
        if not matches:
            if not is_report:
                text = "No artifacts found!" + "\nPossible reasons of this error:\n" + \
                       " * Artifact was not created while building the project due to some internal errors\n" + \
                       " * Artifact path was not specified correctly in 'configs.py'"
                raise CiException(text)

            self.out.log("No artifacts found.")

        for matching_path in matches:
            artifact_name = os.path.basename(matching_path)
            destination = os.path.join(self.artifact_dir, artifact_name)
            if not self.settings.no_archive:
                try:
                    make_big_archive(destination, matching_path)
                    if is_report:
                        artifact_path = self.automation_server.artifact_path(self.artifact_dir, artifact_name + ".zip")
                        self.collected_report_artifacts.add(artifact_path)
                    continue
                except OSError:
                    # Single file archiving is not implemented at the moment
                    pass
            try:
                shutil.copytree(matching_path, destination)
                if is_report:
                    text = "'" + artifact_name + "' is not a file and cannot be reported as an artifact"
                    self.out.log(text)
            except NotADirectoryError:
                shutil.copyfile(matching_path, destination)
                if is_report:
                    artifact_path = self.automation_server.artifact_path(self.artifact_dir, artifact_name)
                    self.collected_report_artifacts.add(artifact_path)

    def collect_step_artifacts(self, step_artifacts, step_report_artifacts):
        if step_artifacts:
            path = utils.parse_path(step_artifacts, self.settings.project_root)
            self.move_artifact(path, is_report=False)
        if step_report_artifacts:
            path = utils.parse_path(step_report_artifacts, self.settings.project_root)
            self.move_artifact(path, is_report=True)

    def report_artifacts(self):
        self.reporter.report_artifacts(list(self.collected_report_artifacts))

    def clean_artifacts_silently(self):
        try:
            shutil.rmtree(self.artifact_dir)
        except OSError:
            pass
        os.makedirs(self.artifact_dir)
        self.html_output.artifact_dir_ready = True

    @staticmethod
    def _check_artifact_absent(artifact_path: str):
        if os.path.exists(artifact_path):
            text: str = f"Build artifact '{os.path.basename(artifact_path)}' already present in artifact directory."
            text += "\nPossible reason of this error: previous build results in working directory"
            raise CriticalCiException(text)



from typing import Union
from requests import Response
from ...lib.gravity import Dependency, Module
from ...lib import utils
from ...lib.module_arguments import ModuleArgumentParser
from .base_server import BaseServerForHostingBuild, BaseServerForTrigger
from .jenkins_server import JenkinsServerForHostingBuild, JenkinsServerForTrigger
from .local_server import LocalServer
from .teamcity_server import TeamcityServer
from .github_server import GithubServer

__all__ = [
    "AutomationServerForHostingBuild",
    "AutomationServerForTrigger"
]


class AutomationServer(Module):

    @staticmethod
    def define_arguments(argument_parser: ModuleArgumentParser) -> None:
        parser = argument_parser.get_or_create_group("Automation server", "Automation server options")
        parser.add_argument("--server-type", "-st", dest="type", choices=["tc", "jenkins", "local"],
                            help="Type of environment to refer to (tc - TeamCity, jenkins - Jenkins, "
                                 "local - user local terminal). TeamCity and Jenkins environment "
                                 "is detected automatically when launched on build agent")


class AutomationServerForHostingBuild(AutomationServer):
    teamcity_driver_factory = Dependency(TeamcityServer)
    local_driver_factory = Dependency(LocalServer)
    jenkins_driver_factory = Dependency(JenkinsServerForHostingBuild)
    github_driver_factory = Dependency(GithubServer)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.driver: BaseServerForHostingBuild = \
            utils.create_driver(teamcity_factory=self.teamcity_driver_factory,
                                jenkins_factory=self.jenkins_driver_factory,
                                github_factory=self.github_driver_factory,
                                local_factory=self.local_driver_factory,
                                env_type=self.settings.type)

    def report_build_location(self) -> str:
        return self.driver.report_build_location()

    def artifact_path(self, local_artifacts_dir: str, item: str) -> str:
        return self.driver.artifact_path(local_artifacts_dir, item)

    def add_build_tag(self, tag: str) -> Response:
        return self.driver.add_build_tag(tag)


class AutomationServerForTrigger(AutomationServer):
    teamcity_driver_factory = Dependency(TeamcityServer)
    local_driver_factory = Dependency(LocalServer)
    jenkins_driver_factory = Dependency(JenkinsServerForTrigger)
    github_driver_factory = Dependency(GithubServer)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.driver: BaseServerForTrigger = \
            utils.create_driver(teamcity_factory=self.teamcity_driver_factory,
                                jenkins_factory=self.jenkins_driver_factory,
                                github_factory=self.github_driver_factory,
                                local_factory=self.local_driver_factory,
                                env_type=self.settings.type)

    def trigger_build(self, revision: str) -> None:
        self.driver.trigger_build(revision)



from .deployment_utils import UniversumRunner
from .utils import python, simple_test_config


def test_minimal_install(clean_docker_main: UniversumRunner):
    # Run without parameters
    log = clean_docker_main.environment.assert_unsuccessful_execution(f"{python()} -m universum")
    assert "No module named universum" not in log

    # Run locally
    log = clean_docker_main.run(simple_test_config, force_installed=True)
    assert clean_docker_main.local.repo_file.name in log

    # Run from Git
    clean_docker_main.clean_artifacts()
    log = clean_docker_main.run(simple_test_config, vcs_type="git", force_installed=True)
    assert clean_docker_main.git.repo_file.name in log

    # Run from P4
    clean_docker_main.clean_artifacts()
    log = clean_docker_main.run(simple_test_config, vcs_type="p4", force_installed=True)
    assert clean_docker_main.perforce.repo_file.name in log


def test_minimal_install_with_git_only(clean_docker_main_no_p4: UniversumRunner, capsys):
    # Run from P4
    clean_docker_main_no_p4.run(simple_test_config, vcs_type="p4", force_installed=True, expected_to_fail=True)
    assert "Please refer to `Prerequisites` chapter of project documentation" in capsys.readouterr().out

    # Run from git
    clean_docker_main_no_p4.clean_artifacts()
    log = clean_docker_main_no_p4.run(simple_test_config, vcs_type="git", force_installed=True)
    assert clean_docker_main_no_p4.git.repo_file.name in log


def test_minimal_install_plain_ubuntu(clean_docker_main_no_vcs: UniversumRunner, capsys):
    # Run from P4
    clean_docker_main_no_vcs.run(simple_test_config, vcs_type="p4", force_installed=True, expected_to_fail=True)
    assert "Please refer to `Prerequisites` chapter of project documentation" in capsys.readouterr().out

    # Run from Git
    clean_docker_main_no_vcs.run(simple_test_config, vcs_type="git", force_installed=True, expected_to_fail=True)
    assert "Please refer to `Prerequisites` chapter of project documentation" in capsys.readouterr().out

    # Run locally
    log = clean_docker_main_no_vcs.run(simple_test_config, force_installed=True)
    assert clean_docker_main_no_vcs.local.repo_file.name in log



from typing import Union, List, Optional
import pytest

from universum import __main__
from universum.lib.module_arguments import IncorrectParameterError, ModuleNamespace
from . import utils


def get_match_all(*args):
    """
    :param args: each argument is a pattern to be found in the error text
    :return: regular expression pattern that matches the string only if each argument pattern is found

    Example:

    >>> get_match_all("a", "b")
    '(?is)(?=.*a)(?=.*b)'
    """

    # (?is) means "ignore case" and "dot matches newline character"
    return "(?is)" + "".join(f"(?=.*{arg})" for arg in args)


def create_settings(test_type: str, vcs_type: str) -> ModuleNamespace:
    settings = utils.create_empty_settings(test_type)
    settings.Output.type = "term"
    if test_type == "github-handler":
        settings.GithubHandler.event = "event"
        settings.GithubHandler.payload = "{}"
        settings.GithubHandler.trigger_url = "http://example.com"
        settings.GithubToken.integration_id = "1234"
        settings.GithubToken.key = "/path"
        return settings

    settings.ProjectDirectory.project_root = "project_root"
    settings.Vcs.type = vcs_type

    if test_type == "poll":
        settings.Poll.db_file = "p4poll.json"
    elif test_type == "submit":
        settings.Submit.commit_message = "Test CL"
    elif test_type == "main":
        settings.AutomationServer.type = "tc"
        settings.TeamcityServer.server_url = "https://teamcity"
        settings.TeamcityServer.build_id = "BuildId_167"
        settings.TeamcityServer.configuration_id = "TestConfig"
        settings.TeamcityServer.user_id = "TeamCityUser"
        settings.TeamcityServer.passwd = "TeamCityPassword"

    if vcs_type in ["git", "gerrit", "ghapp", "ghactions"]:
        if test_type == "main" and vcs_type == "ghactions":
            settings.GithubActionsMainVcs.token = "token"
            settings.GithubActionsMainVcs.payload = "{\"repository\": {\"html_url\": \"None\"}," \
                                                    " \"pull_request\": {\"head\": {\"ref\": \"None\"}}}"
        else:
            settings.GitVcs.repo = "ssh://user@127.0.0.1"
            # the refspec is crafted to satisfy requirements of our gerrit module, and others don't care
            settings.GitVcs.refspec = "refs/changes/00/000000/0"
        if test_type == "submit":
            settings.GitSubmitVcs.user = "Testing User"
            settings.GitSubmitVcs.email = "some@email.com"
        elif test_type == "main" and vcs_type == "ghapp":
            settings.GitMainVcs.checkout_id = "HEAD"
            settings.GithubAppMainVcs.check_id = "000000000"
            settings.GithubToken.integration_id = "1234"
            settings.GithubToken.key = "/path"
            settings.GithubTokenWithInstallation.installation_id = "5678"
    elif vcs_type == "none":
        if test_type == "main":
            settings.LocalMainVcs.source_dir = "temp"
    elif vcs_type == "p4":
        settings.PerforceVcs.port = "127.0.0.1:1666"
        settings.PerforceVcs.user = "p4user"
        settings.PerforceVcs.password = "p4password"

        if test_type == "main":
            settings.PerforceMainVcs.client = "p4_disposable_workspace"
            settings.PerforceMainVcs.force_clean = True
            settings.MainVcs.report_to_review = True
            settings.Swarm.review_id = "123"
            settings.Swarm.change = "456"
            settings.Swarm.server_url = "https://swarm"

        if test_type in ("main", "poll",):
            settings.PerforceWithMappings.project_depot_path = "//depot"

        if test_type == "submit":
            settings.PerforceSubmitVcs.client = "p4_client"

    return settings


def parametrize_unset(parameter_name="unset"):
    return pytest.mark.parametrize(parameter_name, [
        lambda x, y, z: setattr(getattr(x, y), z, None),
        lambda x, y, z: setattr(getattr(x, y), z, "")
    ], ids=["none", "empty"])


def assert_incorrect_parameter(settings: ModuleNamespace, *args):
    with pytest.raises(IncorrectParameterError, match=get_match_all(*args)):
        __main__.run(settings)


missing_params = []


def param(test_type: str, module: str, field: str,
          vcs_type: Union[str, List[str]] = "*", error_match: Optional[str] = None) -> None:

    if isinstance(vcs_type, list):
        for specific_vcs_type in vcs_type:
            param(test_type, module, field, specific_vcs_type, error_match)
        return

    if vcs_type == "*":
        param(test_type, module, field, ["p4", "git", "gerrit", "ghapp", "ghactions", "none"], error_match)
        return

    if test_type == "*":
        for specific_test_type in ["main", "submit", "poll"]:
            param(specific_test_type, module, field, vcs_type, error_match)
        return

    if error_match is None:
        error_match = field

    test_id = test_type + "-" + vcs_type + "-" + module + "." + field
    missing_params.append(pytest.param(test_type, module, field, vcs_type, error_match, id=test_id))


param("submit",         "Submit",                       "commit_message",
      vcs_type=["p4", "git", "gerrit", "ghapp", "ghactions"])
param("submit",         "PerforceSubmitVcs",            "client",          vcs_type="p4")
param("main",           "PerforceMainVcs",              "client",          vcs_type="p4")
param("main",           "LocalMainVcs",                 "source_dir",      vcs_type="none")
param("submit",         "GitSubmitVcs",                 "user",
      vcs_type=["git", "gerrit", "ghapp", "ghactions"])
param("submit",         "GitSubmitVcs",                 "email",
      vcs_type=["git", "gerrit", "ghapp", "ghactions"])
param("main",           "TeamcityServer",               "build_id")
param("main",           "TeamcityServer",               "configuration_id")
param("main",           "TeamcityServer",               "server_url",      error_match="TEAMCITY_SERVER")
param("main",           "TeamcityServer",               "user_id",         error_match="TC_USER")
param("main",           "TeamcityServer",               "passwd",          error_match="TC_PASSWD")
param("*",              "PerforceVcs",                  "port",            vcs_type="p4", error_match="perforce port")
param("*",              "PerforceVcs",                  "user",            vcs_type="p4")
param("*",              "PerforceVcs",                  "password",        vcs_type="p4")
param("main",           "Swarm",                        "server_url",      vcs_type="p4", error_match="SWARM_SERVER")
param("main",           "Swarm",                        "review_id",       vcs_type="p4", error_match="review number")
param("main",           "Swarm",                        "change",          vcs_type="p4")
param("*",              "Vcs",                          "type")
param("*",              "GitVcs",                       "repo",            vcs_type=["git", "gerrit", "ghapp"],
      error_match="git repo")
param("main",           "GitVcs",                       "refspec",         vcs_type="gerrit")
param("main",           "GitMainVcs",                   "checkout_id",     vcs_type="ghapp")
param("main",           "GithubAppMainVcs",             "check_id",        vcs_type="ghapp")
param("main",           "GithubActionsMainVcs",         "token",           vcs_type="ghactions")
param("main",           "GithubActionsMainVcs",         "payload",         vcs_type="ghactions")
param("main",           "GithubToken",                  "integration_id",  vcs_type="ghapp",
      error_match="GITHUB_APP_ID")
param("github-handler", "GithubToken",                  "integration_id",  error_match="GITHUB_APP_ID")
param("main",           "GithubToken",                  "key",             vcs_type="ghapp")
param("github-handler", "GithubToken",                  "key")
param("github-handler", "GithubHandler",                "event",           error_match="GITHUB_EVENT")
param("github-handler", "GithubHandler",                "payload",         error_match="GITHUB_PAYLOAD")
param("github-handler", "GithubHandler",                "trigger_url")
param("main",           "GithubTokenWithInstallation",  "installation_id", vcs_type="ghapp",
      error_match="GITHUB_INSTALLATION_ID")


@parametrize_unset()
@pytest.mark.parametrize("test_type, module, field, vcs_type, error_match", missing_params)
def test_missing_params(unset, test_type, module, field, vcs_type, error_match):
    """
    Make sure the exact error message is produced for the empty settings field
    """
    settings = create_settings(test_type, vcs_type)
    unset(settings, module, field)

    assert_incorrect_parameter(settings, error_match)


@pytest.mark.parametrize("test_type, module, field, vcs_type, error_match", missing_params)
def test_missing_params_correct_error(test_type, module, field, vcs_type, error_match):
    """
    Make sure the error message is not produced if the settings field is not empty
    """
    settings = create_settings(test_type, vcs_type)

    # The idea of the test is to remove some other setting and make sure there is no error message for the current
    # one. For each test type (universum launch type), we choose some settings field to be set to the empty string.
    # At the same time we do not set the current settings field to the empty string. This should generate exception
    # with proper error message, which doesn't contain error regarding the current field passed in parameters.
    # However, this doesn't work if the chosen setting is equal to the one passed in parameters. In that case we use
    # some other setting.
    if test_type == "main":
        if module == "TeamcityServer" and field == "build_id":
            settings.TeamcityServer.configuration_id = ""
        else:
            settings.TeamcityServer.build_id = ""
    elif test_type == "submit":
        if module == "Submit" and field == "commit_message":
            settings.Vcs.type = ""
        else:
            settings.Submit.commit_message = ""
    elif test_type == "poll":
        # The jenkins trigger_url is the only common mandatory settings field of the 'poll' launch type.
        # However, this field is not added to the test parametrization list, so there is no need to set some other
        # field to the empty string to check trigger_url. This parameter is checked in separate test.
        settings.AutomationServer.type = "jenkins"
        settings.JenkinsServerForTrigger.trigger_url = ""
    elif test_type == "github-handler":
        if module == "GithubToken" and field == "integration_id":
            settings.GithubToken.key = ""
        else:
            settings.GithubToken.integration_id = ""

    # the regular expression verifies that the string is not located in the error text
    with pytest.raises(IncorrectParameterError, match=f"(?is)^((?!{error_match}).)*$"):
        __main__.run(settings)


mappings_error_match = get_match_all("P4_PATH", "P4_MAPPINGS")


@parametrize_unset("unset_mappings")
@parametrize_unset("unset_depot_path")
def test_missing_both_perforce_mappings_and_depot_path(unset_mappings, unset_depot_path):
    settings = create_settings("main", "p4")
    unset_mappings(settings, "PerforceWithMappings", "mappings")
    unset_depot_path(settings, "PerforceWithMappings", "project_depot_path")

    assert_incorrect_parameter(settings, "P4_PATH", "P4_MAPPINGS")


def test_present_both_perforce_mappings_and_depot_path():
    settings = create_settings("main", "p4")
    settings.PerforceWithMappings.mappings = ["//depot/... /..."]

    assert_incorrect_parameter(settings, "P4_PATH", "P4_MAPPINGS")


@parametrize_unset()
@pytest.mark.parametrize("vcs_type", ["p4", "git", "gerrit"])
def test_missing_jenkins_params(unset, vcs_type):
    """
    Since varying the type of automation server seems like an overkill for the majority of argument checks,
    it is always set to "tc" (TeamCity). This test checks the correctness of checks for jeknins-specific settings.
    """
    settings = create_settings("main", vcs_type)
    settings.AutomationServer.type = "jenkins"
    unset(settings, "JenkinsServerForHostingBuild", "build_url")

    assert_incorrect_parameter(settings, "Jenkins url of the ongoing build")

    settings = create_settings("poll", vcs_type)
    settings.AutomationServer.type = "jenkins"
    unset(settings, "JenkinsServerForTrigger", "trigger_url")

    assert_incorrect_parameter(settings, "Jenkins url for triggering build")

    settings = create_settings("main", vcs_type)
    settings.AutomationServer.type = "tc"
    settings.TeamcityServer.server_url = ""

    # the regular expression verifies that the string is not located in the error text
    with pytest.raises(IncorrectParameterError, match="(?is)^((?!Jenkins url of the ongoing build).)*$"):
        __main__.run(settings)

    settings = create_settings("poll", vcs_type)
    settings.AutomationServer.type = "tc"
    settings.TeamcityServer.server_url = ""

    # the regular expression verifies that the string is not located in the error text
    with pytest.raises(IncorrectParameterError, match="(?is)^((?!Jenkins url for triggering build).)*$"):
        __main__.run(settings)


@pytest.mark.parametrize("vcs_type", ["git", "none"])
def test_code_review_for_unsupported_vcs(vcs_type):
    settings = create_settings("main", vcs_type)
    settings.MainVcs.report_to_review = True

    assert_incorrect_parameter(settings, "code review system")


def test_gerrit_non_ssh_repo():
    settings = create_settings("main", "gerrit")
    settings.GitVcs.repo = "https://gerrit"

    assert_incorrect_parameter(settings, "ssh protocol")


def test_gerrit_wrong_refspec():
    settings = create_settings("main", "gerrit")
    settings.GitVcs.refspec = "no slashes in this string"

    assert_incorrect_parameter(settings, "incorrect format")


def test_gerrit_no_username_in_ssh_repo():
    settings = create_settings("main", "gerrit")
    settings.GitVcs.repo = "ssh://127.0.0.1"

    assert_incorrect_parameter(settings, "user name")


def test_swarm_changelist_incorrect_format():
    settings = create_settings("main", "p4")
    settings.Swarm.change = "123,456"

    assert_incorrect_parameter(settings, "changelist for unshelving is incorrect")


def test_multiple_errors_main_vcs_type_and_build_id():
    settings = create_settings("main", "p4")
    settings.TeamcityServer.build_id = None
    settings.Vcs.type = None

    assert_incorrect_parameter(settings, "id of the build on TeamCity", "repository type")


def test_multiple_errors_main_none_source_dir_and_build_id():
    settings = create_settings("main", "none")
    settings.TeamcityServer.build_id = None
    settings.LocalMainVcs.source_dir = None
    settings.MainVcs.report_to_review = True

    assert_incorrect_parameter(settings, "id of the build on TeamCity", "SOURCE_DIR", "no code review system")


def test_multiple_errors_main_p4_params():
    settings = create_settings("main", "p4")

    settings.PerforceVcs.port = None
    settings.PerforceVcs.user = None
    settings.PerforceVcs.password = None
    settings.PerforceMainVcs.client = None
    settings.PerforceWithMappings.project_depot_path = None
    settings.Swarm.review_id = None
    settings.Swarm.change = None
    settings.Swarm.server_url = None
    settings.TeamcityServer.server_url = None
    settings.TeamcityServer.build_id = None
    settings.TeamcityServer.configuration_id = None
    settings.TeamcityServer.user_id = None
    settings.TeamcityServer.passwd = None

    assert_incorrect_parameter(settings, "port", "user name", "password", "mappings", "workspace",
                               "URL of the Swarm", "Swarm review number",
                               "Swarm changelist for unshelving is not specified", "URL of the TeamCity",
                               "id of the build on TeamCity", "id of the configuration on TeamCity",
                               "id of the TeamCity user", "password of the TeamCity user")

    settings = create_settings("main", "p4")

    settings.PerforceVcs.port = None
    settings.PerforceVcs.user = None
    settings.PerforceVcs.password = None
    settings.PerforceMainVcs.client = None
    settings.PerforceWithMappings.project_depot_path = None
    settings.Swarm.review_id = None
    settings.Swarm.change = "123,456"
    settings.Swarm.server_url = None

    assert_incorrect_parameter(settings, "port", "user name", "password", "mappings", "workspace",
                               "URL of the Swarm", "Swarm review number",
                               "Swarm changelist for unshelving is incorrect")

    settings = create_settings("main", "p4")

    settings.PerforceVcs.port = None
    settings.PerforceVcs.user = None
    settings.PerforceVcs.password = None
    settings.PerforceMainVcs.client = None
    settings.PerforceWithMappings.project_depot_path = None
    settings.AutomationServer.type = "jenkins"
    settings.JenkinsServerForHostingBuild.build_url = None

    assert_incorrect_parameter(settings, "port", "user name", "password", "mappings", "workspace",
                               "Jenkins url of the ongoing build")


def test_multiple_errors_submit_p4_params_and_commit_message():
    settings = create_settings("submit", "p4")
    settings.Submit.commit_message = None
    settings.PerforceVcs.port = None
    settings.PerforceVcs.user = None
    settings.PerforceVcs.password = None
    settings.PerforceSubmitVcs.client = None

    assert_incorrect_parameter(settings, "COMMIT_MESSAGE", "port", "user name", "password", "P4CLIENT")


def test_multiple_errors_main_git_params_and_build_id():
    settings = create_settings("main", "git")
    settings.TeamcityServer.build_id = None
    settings.GitVcs.repo = None
    settings.MainVcs.report_to_review = True

    assert_incorrect_parameter(settings, "id of the build on TeamCity", "repo", "no code review system")


def test_multiple_errors_submit_git_params_commit_message():
    settings = create_settings("submit", "git")
    settings.Submit.commit_message = None
    settings.GitVcs.repo = None
    settings.GitSubmitVcs.user = None
    settings.GitSubmitVcs.email = None

    assert_incorrect_parameter(settings, "COMMIT_MESSAGE", "repo", "git user name", "git user email")


def test_multiple_errors_main_gerrit_repo_and_build_id():
    settings = create_settings("main", "gerrit")
    settings.TeamcityServer.build_id = None
    settings.GitVcs.repo = None

    assert_incorrect_parameter(settings, "id of the build on TeamCity", "repo")

    settings = create_settings("main", "gerrit")
    settings.TeamcityServer.build_id = None
    settings.GitVcs.repo = "http://"

    assert_incorrect_parameter(settings, "id of the build on TeamCity", "ssh protocol")

    settings = create_settings("main", "gerrit")
    settings.TeamcityServer.build_id = None
    settings.GitVcs.repo = "ssh://127.0.0.1"

    assert_incorrect_parameter(settings, "id of the build on TeamCity", "user name for accessing gerrit")


def test_multiple_errors_main_gerrit_refspec_and_build_id():
    settings = create_settings("main", "gerrit")
    settings.TeamcityServer.build_id = None
    settings.GitVcs.refspec = None
    settings.GitMainVcs.checkout_id = "HEAD"

    assert_incorrect_parameter(settings, "id of the build on TeamCity", "Git refspec for gerrit", "git checkout ID")

    settings = create_settings("main", "gerrit")
    settings.TeamcityServer.build_id = None
    settings.GitVcs.refspec = "ABCDEF"
    settings.GitMainVcs.checkout_id = "HEAD"

    assert_incorrect_parameter(settings, "id of the build on TeamCity", "Git refspec for gerrit", "git checkout ID")


def test_multiple_errors_main_github_and_build_id():
    settings = create_settings("main", "ghapp")
    settings.TeamcityServer.build_id = None
    settings.GitVcs.repo = None
    settings.GitMainVcs.checkout_id = None
    settings.GithubToken.integration_id = None
    settings.GithubToken.key = None
    settings.GithubTokenWithInstallation.installation_id = None
    settings.GithubAppMainVcs.check_id = None

    assert_incorrect_parameter(settings, "id of the build on TeamCity", "git repo", "checkout id for github",
                               "GitHub App ID", "GitHub App private key", "GitHub App installation ID",
                               "GitHub Check Run ID")


def test_multiple_errors_githubhandler():
    settings = create_settings("github-handler", "github")

    settings.GithubHandler.event = None
    settings.GithubHandler.payload = None
    settings.GithubHandler.trigger_url = None
    settings.GithubToken.integration_id = None
    settings.GithubToken.key = None

    assert_incorrect_parameter(settings, "GitHub web-hook event", "build trigger URL", "GitHub web-hook payload",
                               "GitHub App ID", "GitHub App private key")


def test_multiple_errors_poll_p4_params_and_jenkins():
    settings = create_settings("poll", "p4")
    settings.PerforceVcs.port = None
    settings.PerforceVcs.user = None
    settings.PerforceVcs.password = None
    settings.PerforceWithMappings.project_depot_path = None
    settings.AutomationServer.type = "jenkins"
    settings.JenkinsServerForTrigger.trigger_url = None

    assert_incorrect_parameter(settings, "port", "user name", "password", "mappings", "Jenkins url for triggering")



from .lib import utils
from .lib.gravity import Dependency
from .lib.utils import make_block
from .modules import vcs
from .modules.error_state import HasErrorState
from .modules.output import HasOutput
from .modules.structure_handler import HasStructure

__all__ = ["Submit"]


class Submit(HasOutput, HasStructure, HasErrorState):
    description = "Submitting module of Universum"
    vcs_factory = Dependency(vcs.SubmitVcs)

    @staticmethod
    def define_arguments(parser):
        parser.add_argument("--create-review", action="store_true", dest="review",
                            help="create deletable review (shelve for P4, temp branch for Git) "
                                 "instead of actual submitting to repo")
        parser.add_argument("--edit-only", action="store_true", dest="edit_only",
                            help="Only submit existing vcs modifications, no adding or deleting")

        parser.add_argument('--commit-message', '-cm', dest='commit_message', metavar="COMMIT_MESSAGE",
                            help="Commit message. Enter in command line directly, store in environment variable "
                                 "or pass a file path to read the message from by starting the value string with "
                                 "'@'. File path can be either absolute or relative. "
                                 "Please note, that when passing a file, it is expected to be in UTF-8 encoding")
        parser.add_argument("--reconcile-list", "-rl", dest="reconcile_list", metavar="RECONCILE_LIST",
                            help="Comma-separated or linebreak-separated list of files or directories "
                                 "to be reconciled for commit (relative paths starting at client root "
                                 "are supported). To use command line directly, add quotes if needed "
                                 "(e.g. ``-rl 'target1, target2'``). To use a file with reconcile list, "
                                 "start a path to file with '@' (e.g. ``-rl @/path/to/file``)."
                                 "Please note, that when passing a file, it's expected to be in UTF-8 "
                                 "encoding")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.commit_message = self.read_and_check_multiline_option("commit_message", """
            Commit message is not specified.

            Please use '--commit-message' option to provide it. If you start the parameter with '@', the
            rest should be the path to a file containing the commit message text. The path can be absolute
            or relative to the project root. You can also store the message in COMMIT_MESSAGE environment variable.
            Please note, that when passing a file, it's expected to be in UTF-8 encoding
            """)

        try:
            self.reconcile_list = utils.unify_argument_list(self.read_multiline_option("reconcile_list").splitlines())
        except AttributeError:
            self.reconcile_list = []

        self.vcs = self.vcs_factory()
        self.client = None

    @make_block("Executing")
    def execute(self):
        change = self.vcs.driver.submit_new_change(self.commit_message,
                                                   self.reconcile_list,
                                                   review=self.settings.review,
                                                   edit_only=self.settings.edit_only)

        if change == 0:
            self.out.log("Nothing to submit")
        elif self.settings.review:
            self.out.log("Review commit " + change + " created")
        else:
            self.out.log("Change " + change + " submitted")

    @make_block("Finalizing", pass_errors=False)
    def finalize(self):
        self.vcs.finalize()



import argparse
import pathlib
from typing import Callable, List, Optional, Tuple

import yaml

from . import utils, diff_utils


def clang_format_argument_parser() -> argparse.ArgumentParser:
    parser = diff_utils.diff_analyzer_argument_parser("Clang-format analyzer", __file__, "clang-format")
    parser.add_argument("--executable", "-e", dest="executable", default="clang-format",
                        help="The name of the clang-format executable, default: clang-format")
    parser.add_argument("--style", dest="style",
                        help="The 'style' parameter of the clang-format. Can be literal 'file' string or "
                             "path to real file. See the clang-format documentation for details.")
    return parser


def _add_style_param_if_present(cmd: List[str], settings: argparse.Namespace) -> None:
    if settings.style:
        cmd.extend(["-style", settings.style])


@utils.sys_exit
@utils.analyzer(clang_format_argument_parser())
def main(settings: argparse.Namespace) -> List[utils.ReportData]:
    settings.name = "clang-format"
    diff_utils.diff_analyzer_common_main(settings)

    html_diff_file_writer: Optional[Callable[[pathlib.Path, List[str], List[str]], None]] = None
    if settings.write_html:
        wrapcolumn, tabsize = _get_wrapcolumn_tabsize(settings)
        html_diff_file_writer = diff_utils.HtmlDiffFileWriter(settings.target_folder, wrapcolumn, tabsize)

    files: List[Tuple[pathlib.Path, pathlib.Path]] = []
    for src_file_absolute, target_file_absolute, _ in utils.get_files_with_absolute_paths(settings):
        files.append((src_file_absolute, target_file_absolute))
        cmd = [settings.executable, src_file_absolute]
        _add_style_param_if_present(cmd, settings)
        output, _ = utils.run_for_output(cmd)
        with open(target_file_absolute, "w", encoding="utf-8") as output_file:
            output_file.write(output)

    return diff_utils.diff_analyzer_output_parser(files, html_diff_file_writer)


def _get_wrapcolumn_tabsize(settings: argparse.Namespace) -> Tuple[int, int]:
    cmd = [settings.executable, "--dump-config"]
    _add_style_param_if_present(cmd, settings)
    output, error = utils.run_for_output(cmd)
    if error:
        raise utils.AnalyzerException(message="clang-format --dump-config failed with the following error output: " +
                                              error)
    try:
        clang_style = yaml.safe_load(output)
        return clang_style["ColumnLimit"], clang_style["IndentWidth"]
    except yaml.YAMLError as parse_error:
        raise utils.AnalyzerException(message="Parsing of clang-format config produced the following error: " +
                                              str(parse_error))
    except KeyError as key_error:
        raise utils.AnalyzerException(message="Cannot find key in yaml during parsing of clang-format config: " +
                                              str(key_error))


if __name__ == "__main__":
    main()



import json
import urllib.parse
import sh

from ...lib.ci_exception import CiException
from ...lib.gravity import Dependency
from ...lib import utils
from ..reporter import ReportObserver, Reporter
from . import git_vcs

__all__ = [
    "GerritMainVcs",
    "GerritSubmitVcs"
]


class GerritVcs(git_vcs.GitVcs):
    """
    This class contains encapsulates Gerrit VCS functions
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reporter = None

        if not getattr(self.settings, "repo", None):
            return

        if not self.settings.repo.startswith("ssh://"):
            self.error("""
            Only ssh access is supported for gerrit.
            
            Please change the git repo to ssh protocol by using '--git-repo' ('-gr') command
            line parameter or by setting GIT_REPO environment variable.

            Please note that the user name should also be included into git repo
            specification for SSH protocol: --git-repo=ssh://<user>@<host>:<port>/<path>
            """)

        parsed_repo = urllib.parse.urlparse(self.settings.repo)
        self.hostname = parsed_repo.hostname

        if not parsed_repo.username:
            self.error("""
            The user name for accessing gerrit is not specified.
            
            The user name should be included into git repo specification for SSH protocol:
            --git-repo=ssh://<user>@<host>:<port>/<path>
            """)
            return

        self.ssh = sh.ssh.bake(parsed_repo.username+"@"+self.hostname, p=parsed_repo.port)
        self.commit_id = None
        self.review = None
        self.review_version = None

    def run_ssh_command(self, line, stdin=None, stdout=None):
        try:
            return self.ssh(line, _in=stdin, _out=stdout)
        except sh.ErrorReturnCode as e:
            text = f"Got exit code {e.exit_code} while executing the following command:\n{e.full_cmd}"
            if e.stderr:
                text += utils.trim_and_convert_to_unicode(e.stderr) + "\n"
            raise CiException(text) from e


class GerritMainVcs(ReportObserver, GerritVcs, git_vcs.GitMainVcs):
    reporter_factory = Dependency(Reporter)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.settings.checkout_id:
            self.error("""
            The git checkout ID is supplied.

            Please use '--git-refspec' ('-grs') instead of checkout ID for gerrit.
            """)

        if not self.check_required_option("refspec", """
                The git refspec for gerrit is not specified.
    
                For gerrit the git refspec defines the branch to download and the review to work
                with. Please specify the refspec by using '--git-refspec' ('-grs') command line
                parameter or by setting GIT_REFSPEC environment variable.

                Usually, it is enough to set refspec to 'refs/changes/<path>'. For example, on a
                TeamCity server it is enough to set GIT_REFSPEC variable to
                %teamcity.build.branch% for the entire project.
                """):
            return

        refspec = self.settings.refspec
        if refspec.startswith("refs/"):
            refspec = refspec[5:]

        if refspec.count("/") < 2:
            self.error("""
            The git refspec for gerrit has incorrect format.
            
            The git refspec for gerrit must contain components, separated by slash:
            /refs/changes/<number>/<change>/<patch set>.
            Those components define change id and patch set number.
            """)

        self.refspec = refspec

    def code_review(self):
        self.reporter = self.reporter_factory()
        self.reporter.subscribe(self)
        return self

    def update_review_version(self):
        if not self.review:
            self.review = self.refspec.split("/")[2]
        if not self.review_version:
            self.review_version = self.refspec.split("/")[3]

    def get_review_link(self):
        self.update_review_version()
        return "https://" + self.hostname + "/#/c/" + self.review + "/"

    def _get_patch_set_description(self, reference):
        request = "gerrit query --current-patch-set --format json " + reference
        response = str(self.run_ssh_command(request))

        # response is expected to consist of two json objects: patch set description and query summary
        # JSONDecoder.raw_decode() decodes first json object and ignores all that follows
        try:
            decoder = json.JSONDecoder()
            result = decoder.raw_decode(response)
            return result[0]
        except (KeyError, ValueError) as e:
            text = "Error parsing gerrit server response. Full response is the following:\n"
            text += response
            raise CiException(text) from e

    def is_latest_version(self):
        self.update_review_version()
        review_description = self._get_patch_set_description("change:" + self.review)
        latest_version = int(review_description["currentPatchSet"]["number"])

        if int(self.review_version) == latest_version:
            return True

        text = "Current review version is " + self.review_version + \
               ", while latest review version is already " + self.review_latest_version
        self.out.log(text)

        return False

    def calculate_file_diff(self):
        review_description = self._get_patch_set_description("commit:" + self.commit_id)
        target_branch = review_description["branch"]
        reference_commit = self.repo.git.merge_base(target_branch, self.commit_id)
        return self._diff_against_reference_commit(str(reference_commit))

    def code_report_to_review(self, report):
        # git show returns string, each file separated by \n,
        # first line consists of commit id and commit comment, so it's skipped
        commit_files = self.repo.git.show("--name-only", "--oneline", self.commit_id).split('\n')[1:]
        stdin = {'comments': {}}
        text = "gerrit review " + self.commit_id + ' --json '
        for path, issues in report.items():
            if path in commit_files:
                stdin['comments'].update({path: issues})
        self.run_ssh_command(text, json.dumps(stdin))

    def report_start(self, report_text):
        text = "gerrit review --message '" + report_text + "' " + self.commit_id
        self.run_ssh_command(text)

    def report_result(self, result, report_text=None, no_vote=False):
        if result:
            vote = "--label Verified=1 "
        else:
            vote = "--label Verified=-1 "

        text = "gerrit review "
        if not no_vote:
            text += vote
        if report_text:
            text += "--message '" + report_text + "' "

        text += self.commit_id
        self.run_ssh_command(text)

    def prepare_repository(self):
        super().prepare_repository()
        self.commit_id = str(self.repo.head.commit)


class GerritSubmitVcs(GerritVcs, git_vcs.GitSubmitVcs):
    def submit_new_change(self, description, file_list, review=False, edit_only=False):
        change = self.git_commit_locally(description, file_list, edit_only=edit_only)

        self.repo.remotes.origin.push(progress=self.logger, refspec="HEAD:refs/for/" + self.refspec)

        if not review:
            text = "gerrit review --submit --code-review +2 --label Verified=1 --message 'This is an automatic vote' "
            text += change
            self.run_ssh_command(text)

        return change



from typing import Dict, List, Optional, TextIO, Tuple, Type, Union
import json
import shutil
import sh

from . import git_vcs, github_app_vcs, gerrit_vcs, github_actions_vcs, perforce_vcs, local_vcs, base_vcs
from .. import artifact_collector
from ..api_support import ApiSupport
from ..error_state import HasErrorState
from ..project_directory import ProjectDirectory
from ..structure_handler import HasStructure
from ...lib import utils
from ...lib.gravity import Dependency
from ...lib.utils import make_block

__all__ = [
    "MainVcs",
    "PollVcs",
    "SubmitVcs"
]


def create_vcs(class_type: Optional[str] = None) -> Type[ProjectDirectory]:
    driver_factory_class: Union[
        Dict[str, Type[base_vcs.BasePollVcs]],
        Dict[str, Type[base_vcs.BaseSubmitVcs]],
        Dict[str, Type[base_vcs.BaseDownloadVcs]]
    ]
    if class_type == "submit":
        driver_factory_class = {
            "none": base_vcs.BaseSubmitVcs,
            "p4": perforce_vcs.PerforceSubmitVcs,
            "git": git_vcs.GitSubmitVcs,
            "gerrit": gerrit_vcs.GerritSubmitVcs,
            "ghapp": git_vcs.GitSubmitVcs,
            "ghactions": git_vcs.GitSubmitVcs
        }
    elif class_type == "poll":
        driver_factory_class = {
            "none": base_vcs.BasePollVcs,
            "p4": perforce_vcs.PerforcePollVcs,
            "git": git_vcs.GitPollVcs,
            "gerrit": git_vcs.GitPollVcs,
            "ghapp": git_vcs.GitPollVcs,
            "ghactions": git_vcs.GitPollVcs
        }
    else:
        driver_factory_class = {
            "none": local_vcs.LocalMainVcs,
            "p4": perforce_vcs.PerforceMainVcs,
            "git": git_vcs.GitMainVcs,
            "gerrit": gerrit_vcs.GerritMainVcs,
            "ghapp": github_app_vcs.GithubAppMainVcs,
            "ghactions": github_actions_vcs.GithubActionsMainVcs
        }

    vcs_types: List[str] = ["none", "p4", "git", "gerrit", "ghapp", "ghactions"]

    class Vcs(ProjectDirectory, HasStructure, HasErrorState):
        local_driver_factory = Dependency(driver_factory_class['none'])
        git_driver_factory = Dependency(driver_factory_class['git'])
        gerrit_driver_factory = Dependency(driver_factory_class['gerrit'])
        ghapp_driver_factory = Dependency(driver_factory_class['ghapp'])
        ghactions_driver_factory = Dependency(driver_factory_class['ghactions'])
        perforce_driver_factory = Dependency(driver_factory_class['p4'])

        @staticmethod
        def define_arguments(argument_parser):
            parser = argument_parser.get_or_create_group("Source files")

            parser.add_argument("--vcs-type", "-vt", dest="type",
                                choices=vcs_types, metavar="VCS_TYPE",
                                help="Select repository type to download sources from: Perforce ('p4'), "
                                     "Git ('git'), Gerrit ('gerrit'), GitHub App ('ghapp'), "
                                     "GitHub Actions ('ghactions'), or a local directory ('none'). "
                                     "Gerrit uses Git parameters. Each VCS type has its own settings.")

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            if not getattr(self.settings, "type", None):
                self.error(f"""
                    The repository (VCS) type is not set.
                     
                    The repository type defines the version control system that is used for
                    performing the requested action. For example, Universum needs to get project
                    source codes for performing Continuous Integration (CI) builds.

                    The following types are supported: {', '.join(vcs_types)}.
                    
                    Each of these types requires supplying its own
                    configuration parameters. At the minimum, the following
                    parameters are required:
                      * "git", "ghapp", "ghactions" and "gerrit" - GIT_REPO (-gr) and GIT_REFSPEC (-grs)
                      * "perforce"                               - P4PORT (-p4p), P4USER (-p4u), P4PASSWD (-p4P)
                      * "none"                                   - SOURCE_DIR (-fsd)
                      
                    Depending on the requested action, additional type-specific parameters are
                    required. For example, P4CLIENT (-p4c) is required for CI builds with perforce.
                    
                    Please specify the VCS type by using '--vcs-type' ('-vt') command-line option or
                    VCS_TYPE environment variable.
                    """)
                return

            if self.settings.type == "none":
                driver_factory = self.local_driver_factory
            elif self.settings.type == "git":
                driver_factory = self.git_driver_factory
            elif self.settings.type == "gerrit":
                driver_factory = self.gerrit_driver_factory
            elif self.settings.type == "ghapp":
                driver_factory = self.ghapp_driver_factory
            elif self.settings.type == "ghactions":
                driver_factory = self.ghactions_driver_factory
            else:
                driver_factory = self.perforce_driver_factory

            self.driver = driver_factory()

        @make_block("Finalizing")
        def finalize(self):
            self.driver.finalize()

    return Vcs


PollVcs: Type[ProjectDirectory] = create_vcs("poll")
SubmitVcs: Type[ProjectDirectory] = create_vcs("submit")


class MainVcs(create_vcs()):  # type: ignore  # https://github.com/python/mypy/issues/2477
    artifacts_factory = Dependency(artifact_collector.ArtifactCollector)
    api_support_factory = Dependency(ApiSupport)
    driver: base_vcs.BaseDownloadVcs

    @staticmethod
    def define_arguments(argument_parser):
        parser = argument_parser.get_or_create_group("Source files")

        parser.add_argument("--report-to-review", action="store_true", dest="report_to_review", default=False,
                            help="Perform test build for code review system (e.g. Gerrit or Swarm).")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.artifacts: artifact_collector.ArtifactCollector = self.artifacts_factory()
        self.api_support: ApiSupport = self.api_support_factory()

        if not self.settings.report_to_review:
            return

        if not getattr(self, "driver", None):
            return

        self.code_review = self.driver.code_review()
        if not self.code_review:
            self.error("""
                There is no code review system associated with current VCS type, but reporting
                to the code review system is requested.
                
                If the reporting to the code review system is requested, the universum updates
                the review request based on the build status. However, the currently chosen VCS
                doesn't have associated code review system.
                
                Please either disable reporting to the code review system by removing
                '--report-to-review' command-line parameter or change the VCS type by using
                '--vcs-type' ('-vt') command-line option or VCS_TYPE environment variable
                """)

    def login(self):
        self.driver.login()

    def is_latest_review_version(self):
        if self.settings.report_to_review:
            return self.code_review.is_latest_version()
        return True

    @make_block("Preparing repository")
    def prepare_repository(self) -> None:
        status_file: TextIO = self.artifacts.create_text_file("REPOSITORY_STATE.txt")

        try:
            self.driver.prepare_repository()
        finally:
            status_file.write(self.driver.get_repo_status())
            try:
                sources_list = utils.trim_and_convert_to_unicode(sh.ls("-lR", self.settings.project_root))
                status_file.write(f"\n\nFile list:\n{sources_list}\n")
            except sh.ErrorReturnCode_2 as e:
                if "No such file or directory" not in str(e):
                    raise
            status_file.close()

        self.calculate_diff_for_api()

    @make_block("Registering file diff for API")
    def calculate_diff_for_api(self) -> None:
        file_diff = self.driver.calculate_file_diff()
        if file_diff is not None:
            file_diff = json.dumps(file_diff, indent=4)
        self.api_support.add_file_diff(file_diff)

    def clean_sources_silently(self):
        try:
            shutil.rmtree(self.settings.project_root)
        except OSError:
            pass

    @make_block("Revert repository")
    def revert_repository(self) -> Optional[List[Tuple[Optional[str], Optional[str], Optional[str]]]]:
        diff = self.driver.copy_cl_files_and_revert()
        return diff
